"""doctest utilities"""

from collections import defaultdict
import doctest
import json
import re
import unittest

import jsondiff
from jsondiff.symbols import (
    delete,
    insert,
)
from six import string_types


class ParseError(Exception):
    """Error during example output parsing."""


def _parse_webtest_testresponse(string):
    """Parse a webtest.TestResponse string into a (status, headers, body)
    tuple.

    >>> from webtest import TestResponse
    >>> r = TestResponse(b'{a: 1}', status=200,
    ...                  headerlist=[('Content-Type', 'application/json')])
    >>> _parse_webtest_testresponse(str(r))
    ('Response: 200 OK', [('Content-Type', 'application/json')], '{a: 1}')
    >>> r = TestResponse(b'hello!', status=200,
    ...                  headerlist=[('this', 'that'),
    ...                              ('Host', 'example.fr')])

    >>> _parse_webtest_testresponse(str(r))
    ('Response: 200 OK', [('Host', 'example.fr'), ('This', 'that')], 'hello!')
    >>> r = TestResponse(status=204)
    >>> _parse_webtest_testresponse(str(r))
    ('Response: 204 No Content', [], '')
    >>> r = TestResponse(status=204, headerlist=[('Host', 'example.fr')])
    >>> _parse_webtest_testresponse(str(r))
    ('Response: 204 No Content', [('Host', 'example.fr')], '')
    >>> _parse_webtest_testresponse(
    ...     'Response: 204 No Content<BLANKLINE><BLANKLINE>')
    ('Response: 204 No Content', [], '')
    >>> _parse_webtest_testresponse('ah')  # doctest: +IGNORE_EXCEPTION_DETAIL
    Traceback (most recent call last):
        ...
    ParseError: not a webtest.TestResponse string
    """
    for newline in ('\n', doctest.BLANKLINE_MARKER):
        try:
            status, rest = string.split(newline, 1)
        except ValueError:
            continue
        else:
            break
    else:
        raise ParseError('not a webtest.TestResponse string')
    if not re.match('Response: \d{3} \w+', status):
        raise ParseError('not a webtest.TestResponse string')
    headers = []
    while re.match('[a-zA-Z\-]+: [^\n]+\n', rest):
        header, rest = rest.split('\n', 1)
        name, value = header.split(':', 1)
        headers.append((name.strip(), value.strip()))
    body = rest.strip()
    while body.endswith(doctest.BLANKLINE_MARKER):
        body = body[:-len(doctest.BLANKLINE_MARKER)]
    return status, headers, body


def jsondiff_filter_ellipsis(diffobj):
    """Remove keys from `diffobj` if there's an ellipsis in wanted value.

    >>> diffobj = {0: ["...", "123"], 1: [2, 3]}
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    {1: [2, 3]}
    >>> diffobj = {0: ["before...after", "before777after"]}
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    {}
    >>> diffobj = {0: ["before...after", "different"]}
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    {0: ['before...after', 'different']}

    We handle nesting well:

    >>> diffobj = {0: {1: {2: ['a ... number', 'a secret number']}},
    ...            1: ['a', 'b']}
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    {1: ['a', 'b']}
    >>> diffobj = {0: [["..."], ["secret"]]}
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    {}

    Diff structure with "insert" (or "delete") action are ignored:

    >>> diffobj = {0: {delete: {'foo': 'bar'}},
    ...            1: ['...', '666']}
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    {0: {delete: {'foo': 'bar'}}}
    >>> diffobj = {insert: [(0, {'foo': 'bar'})]}
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    {insert: [(0, {'foo': 'bar'})]}

    Marshalled values of these actions are also ignored:
    >>> diffobj = {'$insert': [(0, {'foo': 'bar'})]}
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    {'$insert': [(0, {'foo': 'bar'})]}

    If `diffobj` is a list, it probably means that nothing compares in input,
    so we ignore:

    >>> diffobj = [{'this': 1}, {'that': {'this again': 2}}, {'unrelated': 0}]
    >>> jsondiff_filter_ellipsis(diffobj)
    >>> diffobj
    [{'this': 1}, {'that': {'this again': 2}}, {'unrelated': 0}]
    """
    if isinstance(diffobj, list):
        # Nothing compares?
        return
    for key, diff in list(diffobj.items()):
        if key in (insert, delete):
            continue
        if key in ('$insert', '$delete'):
            continue
        if isinstance(diff, dict):
            jsondiff_filter_ellipsis(diff)
            if not diff:
                del diffobj[key]
        elif isinstance(diff, list):
            want, got = diff
            if isinstance(want, list):
                if len(want) != len(got):
                    continue
                for w, g in zip(want, got):
                    if not doctest._ellipsis_match(w, g):
                        break
                else:
                    del diffobj[key]
            elif isinstance(want, string_types):
                if doctest._ellipsis_match(want, got):
                    del diffobj[key]


def _headerlist_to_dict(headerlist):
    """Convert a list of headers tuples into a dict of lists.

    >>> _headerlist_to_dict([('Location', 'www.example.com')])
    {'Location': ['www.example.com']}
    >>> headers = _headerlist_to_dict([('Location', 'there'), ('Foo', 'bar'),
    ...                                ('Foo', 'baz')])
    >>> list(sorted(headers.items()))
    [('Foo', ['bar', 'baz']), ('Location', ['there'])]
    """
    headers = defaultdict(list)
    for name, value in headerlist:
        headers[name].append(value)
    return dict(headers)


def _headerlist_to_string(headerlist):
    """Convert a list of headers tuples into a multiline string.

    >>> _headerlist_to_string([('Location', 'www.example.com')])
    'Location: www.example.com'
    >>> print(_headerlist_to_string([('Location', 'there'),
    ...                              ('Content-Type', 'application/json')]))
    Location: there
    Content-Type: application/json
    """
    return '\n'.join('{}: {}'.format(name, value)
                     for name, value in headerlist)


class WebTestReponseOutputChecker(doctest.OutputChecker):
    """Output checker for webtest.TestResponse objects."""

    def check_output(self, want, got, optionflags):
        # No super() since doctest.OutputChecker is an old-style class.
        if doctest.OutputChecker.check_output(self, want, got, optionflags):
            return True
        try:
            w_status, w_headers, w_body = _parse_webtest_testresponse(want)
            g_status, g_headers, g_body = _parse_webtest_testresponse(got)
        except ParseError:
            return False
        if w_status != g_status:
            return False
        w_headers = _headerlist_to_dict(w_headers)
        g_headers = _headerlist_to_dict(g_headers)
        # Drop extra headers in "got".
        for hname in list(g_headers):
            if hname not in w_headers:
                del g_headers[hname]
        if g_headers != w_headers:
            if optionflags & doctest.ELLIPSIS:
                for hname, wvalues in w_headers.items():
                    try:
                        gvalues = g_headers[hname]
                    except KeyError:
                        return False
                    for wvalue, gvalue in zip(wvalues, gvalues):
                        if not doctest._ellipsis_match(wvalue, gvalue):
                            return False
            else:
                return False
        if not w_body and g_body:
            return False
        elif not g_body and w_body:
            return False
        elif w_body and g_body:
            try:
                json_want = json.loads(w_body)
                json_got = json.loads(g_body)
            except ValueError:
                return False
            diff = self._filtered_jsondiff(json_want, json_got, optionflags)
            if diff:
                return False
        return True

    @staticmethod
    def _filtered_jsondiff(want, got, optionflags, **kwargs):
        diff = jsondiff.diff(want, got, syntax='symmetric', **kwargs)
        if optionflags & doctest.ELLIPSIS:
            if isinstance(want, list) and isinstance(diff, dict):
                for position, itemdiff in list(diff.items()):
                    jsondiff_filter_ellipsis(itemdiff)
                    if not itemdiff:
                        del diff[position]
            else:
                jsondiff_filter_ellipsis(diff)
        return diff

    def output_difference(self, example, got, optionflags):
        want = example.want
        errors = []
        try:
            w_status, w_headers, w_body = _parse_webtest_testresponse(want)
        except ParseError as exc:
            errors.append('In expected: {!s}\n{}'.format(exc, want))
        try:
            g_status, g_headers, g_body = _parse_webtest_testresponse(got)
        except ParseError as exc:
            errors.append('In actual: {!s}\n{}'.format(exc, got))
        if errors:
            return '\n'.join(errors)
        diff = []
        if w_status != g_status:
            diff.append('Expected status: {}'.format(w_status))
            diff.append('Actual status: {}'.format(g_status))
        if w_headers != g_headers:
            diff.append('Expected headers: {}'.format(
                _headerlist_to_string(w_headers)))
            diff.append('Actual headers: {}'.format(
                _headerlist_to_string(g_headers)))
        if w_body or g_body:
            if not w_body:
                diff.append('Expected body is empty')
            elif not g_body:
                diff.append('Actual body is empty')
            else:
                both_valid = True
                try:
                    w_json = json.loads(w_body)
                except ValueError:
                    diff.append(
                        'Expected body is invalid JSON: {}'.format(w_body))
                    both_valid = False
                try:
                    g_json = json.loads(g_body)
                except ValueError:
                    diff.append(
                        'Actual body is invalid JSON: {}'.format(g_body))
                    both_valid = False
                if both_valid:
                    # Marshal the diff to eliminate jsondiff symbols before
                    # dumping.
                    body_diff = self._filtered_jsondiff(
                        w_json, g_json, optionflags, marshal=True)
                    if body_diff:
                        msg = '\n'.join([
                            'JSON bodies differ',
                            'expected\n{}',
                            'actual\n{}',
                            'diff:\n{}'
                        ])
                        diff.append(msg.format(json.dumps(w_json, indent=2),
                                               json.dumps(g_json, indent=2),
                                               json.dumps(body_diff)))
        return '\n'.join(diff)


def setUp(test, cls=unittest.TestCase, globs_from_testcase=()):
    """set-up function for a doctest that uses the setUp method of `cls` class
    (usually a subclass of unittest.TestCase) and insert attributes of the
    test case instance as "globs" of the DocTest object.
    """
    # unittest(2).TestCase.__init__ takes a string corresponding to one of its
    # method, so create a fake one and cleanup afterwards.
    def fake_test_method(testcase):
        pass

    fakename = '_fake_test'
    assert not hasattr(cls, fakename)
    setattr(cls, fakename, fake_test_method)
    try:
        test._tc = cls(fakename)
        test._tc.setUpClass()
        test._tc.setUp()
        for attrname in globs_from_testcase:
            test.globs[attrname] = getattr(test._tc, attrname)
    finally:
        delattr(cls, fakename)


def tearDown(test):
    """tear-down function performing cleanup actions for a DocTest bound to a
    unittest-like TestCase
    """
    if not hasattr(test, '_tc'):
        return
    test._tc.tearDown()
    test._tc.tearDownClass()


if __name__ == '__main__':
    doctest.testmod()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-jsonschema entity's classes"""

from logilab.common.registry import objectify_predicate

from cubicweb.predicates import match_kwargs, non_final_entity
from cubicweb.view import Adapter, EntityAdapter

from cubicweb_jsonschema import (
    VIEW_ROLE,
    CREATION_ROLE,
    EDITION_ROLE,
    orm_rtype,
)


def jsonschema_adapter(cnx, **ctx):
    """Return a IJSONSchema adapter selected from ``ctx`` information."""
    return cnx.vreg['adapters'].select('IJSONSchema', cnx, **ctx)


class IJSONSchemaETypeAdapter(Adapter):
    """IJSONSchema adapter for entity creation with `etype` specified in
    selection context.

    Subclasses may refine the selection context using ``match_kwargs``
    predicates (e.g. ``match_kwargs({'etype': 'MyEtype'})`` to override JSON
    Schema generation for a particular etype).
    """
    __regid__ = 'IJSONSchema'
    __select__ = match_kwargs('etype')

    def __repr__(self):
        return ('<{0.__class__.__name__} etype={0.etype}>'.format(self))

    @property
    def etype(self):
        """The entity type bound to this adapter."""
        return str(self.cw_extra_kwargs['etype'])

    @property
    def entity_mapper(self):
        return self._cw.vreg['mappers'].select(
            'jsonschema.entity', self._cw, etype=self.etype)

    def view_schema(self, **kwargs):
        """Return a JSON schema suitable for viewing adapted entity."""
        jsldoc = self.entity_mapper.jsl_document(VIEW_ROLE)
        return jsldoc.get_schema(**kwargs)

    def creation_schema(self, **kwargs):
        """Return a JSON schema suitable for entity creation."""
        jsldoc = self.entity_mapper.jsl_document(CREATION_ROLE)
        return jsldoc.get_schema(**kwargs)

    @property
    def entity(self):
        return self._cw.vreg['etypes'].etype_class(self.etype)(self._cw)

    def create_entity(self, instance):
        """Return a CubicWeb entity built from `instance` data matching this
        JSON schema.
        """
        values = self.entity_mapper.values(instance)
        return self._cw.create_entity(self.etype, **values)

    def serialize(self, entities):
        """Return `entities` as a collection for JSON serialization."""
        mapper = self._cw.vreg['mappers'].select(
            'jsonschema.collection', self._cw, etype=self.etype)
        return mapper.serialize(entities)


class IJSONSchemaRelationTargetETypeAdapter(IJSONSchemaETypeAdapter):
    """IJSONSchema adapter for entity creation of an entity of `etype` related
    through `rtype` and `role` specified in selection context.
    """
    __select__ = match_kwargs('etype', 'rtype', 'role')

    def __repr__(self):
        return ('<{0.__class__.__name__} etype={0.etype} rtype={0.rtype} '
                'role={0.role}>'.format(self))

    @property
    def rtype(self):
        return self.cw_extra_kwargs['rtype']

    @property
    def role(self):
        return self.cw_extra_kwargs['role']

    @property
    def entity_mapper(self):
        return self._cw.vreg['mappers'].select(
            'jsonschema.entity', self._cw,
            etype=self.etype, rtype=self.rtype, role=self.role,
        )

    def create_entity(self, instance, target=None):
        """Return a CubicWeb entity related to `target` through relation
        information from selection context.
        """
        entity = super(IJSONSchemaRelationTargetETypeAdapter,
                       self).create_entity(instance)
        if target is not None:
            entity.cw_set(**{orm_rtype(self.rtype, self.role): target})
        return entity


class IJSONSchemaEntityAdapter(EntityAdapter):
    """IJSONSchema adapter for live entities."""
    __regid__ = 'IJSONSchema'
    # Prevent this adapter from being selected in place of
    # IJSONSchemaRelatedEntityAdapter in case one picks a bad rtype/role.
    __select__ = non_final_entity() & ~match_kwargs('rtype', 'role')

    def __repr__(self):
        return ('<{0.__class__.__name__} entity={0.entity}>'.format(self))

    @property
    def etype(self):
        return self.entity.cw_etype

    @property
    def entity_mapper(self):
        return self._cw.vreg['mappers'].select(
            'jsonschema.entity', self._cw, entity=self.entity)

    def view_schema(self, **kwargs):
        """Return a JSON schema suitable for viewing adapted entity."""
        jsldoc = self.entity_mapper.jsl_document(VIEW_ROLE)
        return jsldoc.get_schema(**kwargs)

    def edition_schema(self, **kwargs):
        """Return a JSON schema suitable for editing adapted entity."""
        jsldoc = self.entity_mapper.jsl_document(EDITION_ROLE)
        return jsldoc.get_schema(**kwargs)

    def relations(self):
        """Yield (rtype, role, targettypes) for non-final relations."""
        rsection = self._cw.vreg['uicfg'].select(
            'jsonschema', self._cw, entity=self.entity)
        return rsection.relations_by_section(self.entity, 'related', 'read')

    def edit_entity(self, instance):
        """Return a CubicWeb entity built from `instance` data matching this
        JSON schema.
        """
        values = self.entity_mapper.values(instance)
        return self.entity.cw_set(**values)

    def serialize(self):
        """Return a dictionary of entity's data suitable for JSON
        serialization.
        """
        return self.entity_mapper.serialize()


@objectify_predicate
def relation_for_entity(cls, cnx, entity, rtype, role):
    """Return 1 if `entity` has an `(rtype, role)` relation."""
    return 1 if entity.e_schema.has_relation(rtype, role) else 0


class IJSONSchemaRelatedEntityAdapter(IJSONSchemaEntityAdapter):
    """IJSONSchema adapter for entities in the context of a relation."""
    __select__ = (non_final_entity()
                  & match_kwargs('rtype', 'role') & relation_for_entity())

    def __repr__(self):
        return ('<{0.__class__.__name__} entity={0.entity} '
                'rtype={0.rtype} role={0.role}>'.format(self))

    @property
    def rtype(self):
        return self.cw_extra_kwargs['rtype']

    @property
    def role(self):
        return self.cw_extra_kwargs['role']



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""Appobjects for mapping Yams schema definitions to JSON Schema documents."""

from collections import OrderedDict
from functools import partial

import iso8601
from six import text_type, string_types

import jsl

from logilab.common.decorators import cachedproperty, classproperty
from logilab.common.registry import Predicate, objectify_predicate, yes

from yams import BadSchemaDefinition, ValidationError
from yams.constraints import StaticVocabularyConstraint
from cubicweb import Binary, neg_role, _
from cubicweb.appobject import AppObject
from cubicweb.predicates import (
    ExpectedValuePredicate,
    PartialPredicateMixIn,
    match_kwargs,
)

from cubicweb_jsonschema import (
    CREATION_ROLE,
    EDITION_ROLE,
    VIEW_ROLE,
    orm_rtype,
)
from cubicweb_jsonschema.views import jsonschema_section


def _etype_from_context(kwargs):
    etype = kwargs.get('etype')
    if etype is None:
        entity = kwargs.get('entity')
        if entity is not None:
            etype = entity.cw_etype
    return etype


@objectify_predicate
def yams_final_rtype(cls, cnx, rtype, role, **kwargs):
    """Predicate that returns 1 when the supplied `rtype` is not final."""
    etype = _etype_from_context(kwargs)
    if etype is None:
        return 0
    rdef = cnx.vreg.schema[etype].rdef(rtype, role=role, takefirst=True)
    if rdef.final:
        return 1
    return 0


@objectify_predicate
def yams_component_target(cls, cnx, rtype, role, target_types=None, **kwargs):
    """Predicate that returns 1 when the target entity types are component of
    the relation defined by `rtype` and `role` (i.e. the relation has
    composite set to `role`).
    """
    etype = _etype_from_context(kwargs)
    if etype is None:
        return 0
    component = None
    eschema = cnx.vreg.schema[etype]
    if target_types is None:
        rschema = cnx.vreg.schema[rtype]
        target_types = tuple(t.type for t in rschema.targets(etype, role=role))
    for target_type in target_types:
        rdef = eschema.rdef(rtype, role=role, targettype=target_type)
        _component = rdef.composite == role
        if component is None:
            component = _component
        elif not component == _component:
            cls.warning('component inconsistency accross target types')
            return 0
    return component


class yams_match(Predicate):
    """Predicate that returns a positive value when the supplied relation
    match parameters given as predicate argument.

    The more __call__ arguments match __init__'s ones, the higher the score
    is.
    """

    def __init__(self, etype=None, rtype=None, role=None, target_types=()):
        self.etype = etype
        self.rtype = rtype
        self.role = role
        if isinstance(target_types, string_types):
            target_types = {target_types}
        elif not isinstance(target_types, (set, frozenset)):
            target_types = frozenset(target_types)
        self.target_types = target_types

    def __call__(self, cls, cnx, rtype, role, **kwargs):
        etype = _etype_from_context(kwargs)
        score = 0
        for key in ('etype', 'rtype', 'role'):
            expected = getattr(self, key, None)
            if expected is not None:
                if locals()[key] != expected:
                    return 0
                score += 1
        target_types = kwargs.get('target_types', frozenset())
        assert isinstance(target_types, (set, frozenset)), \
            'yams_match must be called with a set as "target_types" argument'
        if self.target_types:
            if not target_types.issubset(self.target_types):
                return 0
            score += 1
        return score


class partial_yams_match(PartialPredicateMixIn, ExpectedValuePredicate):
    """Return non-zero if selected class's attributes `etype`, `rtype`, `role`
    and/or `target_types` match with keyword arguments from context.
    """

    match_attributes = ('etype', 'rtype', 'role', 'target_types')

    def __init__(self, *expected, **kwargs):
        super(partial_yams_match, self).__init__((), **kwargs)

    def complete(self, cls):
        self.expected = {}
        for name in self.match_attributes:
            if hasattr(cls, name):
                self.expected[name] = set([getattr(cls, name)])

    def _values_set(self, cls, req, **kwargs):
        values = {}
        for name in self.match_attributes:
            if name in kwargs:
                values[name] = kwargs[name]
        return values


def base_fields(etype):
    fields = OrderedDict()
    fields['__etype__'] = jsl.fields.StringField(
        required=True, enum=[etype], default=etype)
    fields['__eid__'] = jsl.fields.StringField(required=True)
    return fields


def default_options(_cw, etype):
    """Return an Options class for a jsl.Document named `etype`."""
    attrs = {
        'definition_id': etype,
        'title': _cw._(etype),
        'schema_uri': 'http://json-schema.org/draft-06/schema#',
    }
    return type('Options', (object,), attrs)


class JSONSchemaMapper(AppObject):
    """Base class for mappers between Yams schema definitions and JSON Schema
    documents.
    """
    __registry__ = 'mappers'


class ETypeMapper(JSONSchemaMapper):
    """A mapper to build JSL document for an entity type."""
    __regid__ = 'jsonschema.entity'
    __select__ = match_kwargs('etype')

    # Should "base" fields __etype__/__eid__ be included in main schema
    # document?
    include_base_fields = False

    @property
    def etype(self):
        return self.cw_extra_kwargs['etype']

    @property
    def entity(self):
        """Entity bound to this mapper, None for ETypeMapper."""
        return None

    def jsl_document(self, schema_role=None):
        """Return a jsl.Document for bound entity type with fields relations
        registered to appear in the document (through `jsonschema_section`).
        """
        attrs = OrderedDict()
        if self.include_base_fields:
            attrs.update(base_fields(self.etype))
        for mapper in self._object_mappers():
            attrs[mapper.name] = mapper.jsl_field(schema_role)
        for rtype, role, target_types in self.relations(schema_role):
            if target_types is None:
                target_types = self._rtype_target_types(rtype, role)
            assert isinstance(target_types, set), target_types
            mapper = self._relation_mapper(rtype, role, target_types)
            attrs[rtype] = mapper.jsl_field(schema_role)
        attrs['Options'] = default_options(self._cw, self.etype)
        return type(str(self.etype), (jsl.Document, ), attrs)

    def values(self, instance):
        """Return a dict with deserialized data from `instance` suitable for
        insertion in CubicWeb database.
        """
        entity = self.entity
        values = {}
        # Deserialize "jsonschema.object" mappers first.
        for mapper in self._object_mappers():
            values.update(mapper.values(entity, instance))
        # Then Yams relations.
        if entity is None:
            schema_role = CREATION_ROLE
        else:
            schema_role = EDITION_ROLE
        for rtype, role, target_types in self.relations(schema_role):
            mapper = self._relation_mapper(rtype, role, target_types)
            values.update(mapper.values(entity, instance))
        if instance:
            # All properties in "instance" should have been consumed at this
            # point.
            eid = entity.eid if entity is not None else None
            msg = _('unexpected properties: {}').format(', '.join(instance))
            raise ValidationError(eid, {None: msg})
        return values

    def relations(self, schema_role):
        """Yield relation information tuple (rtype, role, targettypes)
        for given schema role in the context of bound entity type.
        """
        try:
            permission = {
                VIEW_ROLE: 'read',
                CREATION_ROLE: 'add',
                EDITION_ROLE: 'update',
            }[schema_role]
        except KeyError:
            raise ValueError('unhandled schema role "{0}" in {1}'.format(
                schema_role, self))
        entity = self.entity
        if entity is None:
            entity = self._cw.vreg['etypes'].etype_class(self.etype)(self._cw)
        rsection = self._cw.vreg['uicfg'].select(
            'jsonschema', self._cw, entity=entity)
        return rsection.relations_by_section(entity, 'inlined', permission)

    def _object_mappers(self):
        """Yield 'jsonschema.object' mapper instance selectable for entity
        bound to this mapper.
        """
        if 'jsonschema.object' not in self._cw.vreg['mappers']:
            return
        for mappercls in self._cw.vreg['mappers']['jsonschema.object']:
            if mappercls.__select__(mappercls, self._cw, etype=self.etype) > 0:
                yield mappercls(self._cw)

    def _rtype_target_types(self, rtype, role):
        rschema = self._cw.vreg.schema[rtype]
        return {t.type for t in rschema.targets(self.etype, role)}

    def _relation_mapper(self, rtype, role, target_types=None):
        return self._cw.vreg['mappers'].select(
            'jsonschema.relation', self._cw,
            etype=self.etype, rtype=rtype, role=role, target_types=target_types)


class TargetETypeMapper(ETypeMapper):
    """Specialized version of :class:`ETypeMapper` selectable for an entity
    type as target of (`rtype`, `role`) relation.
    """
    __select__ = match_kwargs('etype', 'rtype', 'role')

    @property
    def rtype(self):
        return self.cw_extra_kwargs['rtype']

    @property
    def role(self):
        return self.cw_extra_kwargs['role']

    def relations(self, schema_role):
        relations = super(TargetETypeMapper, self).relations(schema_role)
        for rtype, role, target_types in relations:
            if (rtype, role) == (self.rtype, self.role):
                continue
            yield rtype, role, target_types


class EntityMapper(ETypeMapper):
    """Specialized version of :class:`ETypeMapper` for a live entity."""

    __select__ = match_kwargs('entity')

    @property
    def entity(self):
        """Live entity from selection context."""
        return self.cw_extra_kwargs['entity']

    @property
    def etype(self):
        return self.entity.cw_etype

    def _relation_mapper(self, rtype, role, target_types=None):
        return self._cw.vreg['mappers'].select(
            'jsonschema.relation', self._cw,
            entity=self.entity, rtype=rtype,
            role=role, target_types=target_types)

    def serialize(self):
        """Return the serialized value entity bound to this mapper."""
        entity = self.entity
        entity.complete()
        data = {}
        for mapper in self._object_mappers():
            data[mapper.name] = mapper.serialize(entity)
        for rtype, role, target_types in self.relations(VIEW_ROLE):
            relation_mapper = self._relation_mapper(rtype, role, target_types)
            value = relation_mapper.serialize(entity)
            if value is None:
                continue
            data[relation_mapper.orm_rtype] = value
        return data


class TargetEntityMapper(EntityMapper, TargetETypeMapper):
    """Specialized version of :class:`TargetETypeMapper` and
    :class:`EntityMapper` selectable for an *entity* as target of (`rtype`,
    `role`) relation.
    """
    __select__ = match_kwargs('entity', 'rtype', 'role')


class EntityCollectionMapper(JSONSchemaMapper):
    """Mapper for a collection of entities."""
    __regid__ = 'jsonschema.collection'
    __select__ = match_kwargs('etype')

    @property
    def etype(self):
        return self.cw_extra_kwargs['etype']

    @property
    def title(self):
        """Title of the collection, plural form of entity type."""
        return self._cw._('{}_plural').format(self.etype)

    def jsl_field(self, schema_role, **kwargs):
        """Return a JSON schema dict to model a collection of entities."""
        item_mapper = self._cw.vreg['mappers'].select(
            'jsonschema.item', self._cw)
        item_field = item_mapper.jsl_field(schema_role, **kwargs)
        return jsl.fields.ArrayField(title=self.title, items=item_field)

    def serialize(self, entities):
        """Return a list of collection item representing each entity in
        `entities`.
        """
        mapper = self._cw.vreg['mappers'].select(
            'jsonschema.item', self._cw)
        return [mapper.serialize(entity) for entity in entities]


class RelatedCollectionMapper(EntityCollectionMapper):
    """Mapper for a collection of entities through a relation."""
    __select__ = match_kwargs('rtype', 'role')

    @property
    def title(self):
        """Title of the collection, name of the relation."""
        rtype = self.cw_extra_kwargs['rtype']
        role = self.cw_extra_kwargs['role']
        return self._cw._(rtype if role == 'subject' else rtype + '-object')


class CollectionItemMapper(JSONSchemaMapper):
    """Mapper for an item of a collection."""
    __regid__ = 'jsonschema.item'
    __select__ = yes()

    def jsl_field(self, schema_role, **kwargs):
        """Return a jsl.fields.DictField with item description."""
        if schema_role == CREATION_ROLE:
            return jsl.fields.StringField(**kwargs)
        return jsl.fields.DictField(
            properties={
                'type': jsl.fields.StringField(),
                'id': jsl.fields.StringField(),
                'title': jsl.fields.StringField(),
            },
            **kwargs
        )

    @staticmethod
    def serialize(entity):
        """Return a dictionary with entity represented as a collection item."""
        return {
            'type': entity.cw_etype.lower(),
            'id': text_type(entity.eid),
            'title': entity.dc_title(),
        }


class BaseRelationMapper(JSONSchemaMapper):
    """Base abstract class to fill the gap between a yams relation and it's json
    schema mapping.

    They should be selected depending on the relation (`etype`, `rtype`, `role`
    and optionaly `target_types`). You may then get a Jsl field from them by
    calling the `jsl_field` method.
    """
    __regid__ = 'jsonschema.relation'
    __select__ = match_kwargs('rtype', 'role')
    __abstract__ = True

    @property
    def jsl_field_class(self):
        """The jsl.fields.BaseField child class to be returned by the factory.
        """
        raise NotImplementedError

    @property
    def etype(self):
        """The entity type bound to this mapper."""
        return _etype_from_context(self.cw_extra_kwargs)

    def __init__(self, _cw, **kwargs):
        #: relation type name
        self.rtype = kwargs.pop('rtype')
        #: role of `etype` in relation
        self.role = kwargs.pop('role')
        #: possible target types of the relation (empty for attribute relations)
        self.target_types = []
        rschema = _cw.vreg.schema[self.rtype]
        target_types = kwargs.pop('target_types', None)
        for target_eschema in sorted(rschema.targets(role=self.role)):
            if target_eschema.final:
                continue
            target_type = target_eschema.type
            if target_types is not None and target_type not in target_types:
                continue
            self.target_types.append(target_type)

        super(BaseRelationMapper, self).__init__(_cw, **kwargs)

    def __repr__(self):
        return ('<{0.__class__.__name__} etype={0.etype} rtype={0.rtype} '
                'role={0.role} target_types={0.target_types}>'.format(self))

    @cachedproperty
    def description(self):
        ntargets = len(self.target_types)
        if ntargets > 1:
            return None
        elif ntargets == 1:
            targettype = self.target_types[0]
        else:
            targettype = None
        eschema = self._cw.vreg.schema[self.etype]
        rdef = eschema.rdef(self.rtype, role=self.role,
                            targettype=targettype)
        if rdef.description:
            return self._cw._(rdef.description)

    @cachedproperty
    def title(self):
        if self.role == 'object':
            return self._cw._(self.rtype + '_object')
        return self._cw._(self.rtype)

    @cachedproperty
    def orm_rtype(self):
        return orm_rtype(self.rtype, self.role)

    def jsl_field(self, schema_role, **kwargs):
        """Return a JSL field built from `jsl_field_class`."""
        if 'title' not in kwargs:
            kwargs['title'] = self.title
        if 'description' not in kwargs and self.description is not None:
            kwargs['description'] = self.description
        kwargs.pop('module', None)
        return self.jsl_field_class(**kwargs)

    def values(self, entity, instance):
        """Return a dictionary holding deserialized value for this field,
        given the input entity (None on creation) and `instance` (JSON values
        as dictionary).

        If present, the field matching mapped `rtype` is removed from
        `instance` after processing.
        """
        raise NotImplementedError()

    def serialize(self, entity):
        """Return the serialized value for this field."""
        raise NotImplementedError()

    @staticmethod
    def _type(json_value):
        """Return properly typed value for use within a cubicweb's entity from
        given JSON value.

        Nothing to do by default.
        """
        return json_value


class AttributeMapper(BaseRelationMapper):
    """Base abstract class to map attribute relation."""
    __abstract__ = True
    __select__ = yams_final_rtype() & match_kwargs('etype')

    #: JSON Schema "format" keyword for semantic validation.
    format = None

    @property
    def attr(self):
        """Relation definition for bound attribute."""
        return self._cw.vreg.schema[self.etype].rdef(self.rtype)

    def add_constraint(self, cstr, jsl_attrs):
        serializer = self._cw.vreg['mappers'].select_or_none(
            'jsonschema.constraint', self._cw.vreg, self._cw._,
            self.etype, self.rtype, cstr, jsl_attrs)
        if serializer is not None:
            new_attrs = serializer.todict()
            if new_attrs:
                jsl_attrs.update(new_attrs)
        elif not isinstance(cstr, StaticVocabularyConstraint):
            self.warning('JSL: ignored %s on %s', cstr.type(), self.attr)

    def jsl_field(self, schema_role, **kwargs):
        kwargs.setdefault('format', self.format)
        if schema_role in (CREATION_ROLE, EDITION_ROLE):
            if 'required' not in kwargs and self.attr.cardinality[0] == '1':
                kwargs['required'] = True
            if 'default' not in kwargs and self.attr.default is not None:
                kwargs['default'] = self.attr.default
            vocabulary_constraint = next(
                (cstr for cstr in self.attr.constraints
                 if isinstance(cstr, StaticVocabularyConstraint)), None)
            if vocabulary_constraint:
                # In presence of a vocabulary constraint, we wrap the field
                # into a oneOf field with a single-value 'enum', ignoring
                # other constraints.
                fields = []
                field_factory = super(AttributeMapper, self).jsl_field
                for v in sorted(vocabulary_constraint.vocabulary()):
                    kw = {'enum': [v], 'title': self._cw._(v)}
                    fields.append(field_factory(schema_role, **kw))
                return jsl.fields.OneOfField(fields, **kwargs)
            for constraint in self.attr.constraints:
                self.add_constraint(constraint, kwargs)
        return super(AttributeMapper, self).jsl_field(schema_role, **kwargs)

    def values(self, entity, instance):
        """Return a dictionary holding deserialized value for mapped attribute.

        If mapped attribute is absent from `instance` and `entity` is not
        None, the default value for attribute is returned as value of the
        dictionnary.
        """
        try:
            value = instance.pop(self.rtype)
        except KeyError:
            if entity is None:
                return {}
            rschema = entity.e_schema.rdef(self.rtype, self.role)
            value = rschema.default
        else:
            value = self._type(value)
            if entity is not None and getattr(entity, self.orm_rtype) == value:
                # Do not trigger update, if the value has not changed.
                # This is useful for PUT requests for which all fields should
                # be present even if they are not changed. By skipping
                # unchanged value, we avoid possible security check that would
                # be meaningless since the value is not changed.
                return {}
        return {self.orm_rtype: value}

    def serialize(self, entity):
        value = getattr(entity, self.orm_rtype)
        if value is not None:
            return self._value(value)

    @staticmethod
    def _value(value):
        """Return the serializable value from attribute `value`."""
        return value


class ChoiceStringField(jsl.fields.StringField):
    """A string field with choices described in a oneOf array with items made
    of {'enum': [<choice value>], 'title': <choice title>} objects.
    """

    def __init__(self, choices, **kwargs):
        super(ChoiceStringField, self).__init__(**kwargs)
        self.choices = choices

    def get_definitions_and_schema(self, **kwargs):
        definitions, schema = super(
            ChoiceStringField, self).get_definitions_and_schema(**kwargs)
        schema['oneOf'] = [{'enum': [value], 'title': title}
                           for value, title in self.choices]
        return definitions, schema


class StringMapper(AttributeMapper):
    """Attribute mapper for Yams' String type."""
    __select__ = yams_match(target_types='String')
    jsl_field_class = jsl.fields.StringField
    _type = text_type


class FloatMapper(AttributeMapper):
    """Attribute mapper for Yams' Float type."""
    __select__ = yams_match(target_types='Float')
    jsl_field_class = jsl.fields.NumberField


class IntMapper(AttributeMapper):
    """Attribute mapper for Yams' Int type."""
    __select__ = yams_match(target_types='Int')
    jsl_field_class = jsl.fields.IntField


class BooleanMapper(AttributeMapper):
    """Attribute mapper for Yams' Boolean type."""
    __select__ = yams_match(target_types='Boolean')
    jsl_field_class = jsl.fields.BooleanField


class PasswordMapper(AttributeMapper):
    """Attribute mapper for Yams' Password type."""
    __select__ = yams_match(target_types='Password')
    jsl_field_class = jsl.fields.StringField
    #:
    format = 'password'

    def jsl_field(self, schema_role, **kwargs):
        if schema_role == EDITION_ROLE:
            kwargs.setdefault('required', False)
        return super(PasswordMapper, self).jsl_field(schema_role, **kwargs)

    def values(self, entity, instance):
        password_changed = self.orm_rtype in instance
        values = super(PasswordMapper, self).values(entity, instance)
        if entity is not None and not password_changed:
            # We don't want the Password value to be reset if it has not
            # changed.
            del values[self.orm_rtype]
        return values

    def serialize(self, entity):
        return None

    @staticmethod
    def _type(json_value):
        """Return an encoded string suitable for Password type."""
        return json_value.encode('utf-8')


class DateMapper(StringMapper):
    """Attribute mapper for Yams' Date type."""
    __select__ = yams_match(target_types=('Date'))
    #:
    format = 'date'

    @staticmethod
    def _type(value):
        """Return a datetime object parsed from ISO8601 `value` string."""
        return iso8601.parse_date(value)


class DatetimeMapper(DateMapper):
    """Attribute mapper for Yams' (TZ)Datetime type."""
    __select__ = yams_match(target_types=('Datetime', 'TZDatetime'))
    #:
    format = 'date-time'


class BytesMapper(StringMapper):
    """Attribute mapper for Yams' Bytes type."""
    __select__ = yams_match(target_types='Bytes')

    @staticmethod
    def _type(value):
        """Return a Binary containing `value`."""
        return Binary(value.encode('utf-8'))

    @staticmethod
    def _value(value):
        """Return a unicode string from Binary `value`."""
        return value.getvalue().decode('utf-8')


class CompoundMapper(JSONSchemaMapper):
    """Mapper for a "compound" field gathering Yams relations into a
    dedicated JSON "object" to be inserted in "definitions" key of the JSON
    Schema document.

    The compound "object" will appear in the main JSON Schema document's
    properties under the `name` class attribute (defaults to class name).
    """
    __regid__ = 'jsonschema.object'
    __abstract__ = True
    __select__ = partial_yams_match()

    #: entity type holding this compound field.
    etype = None
    #: sequence of relations gathered in this compound field
    relations = ()
    #: title of the field
    title = None

    # sequence of 'relation/role' computed from 'relations'
    _relations = ()

    @classproperty
    def name(cls):
        """name of the property to be inserted in the main JSON Schema
        document (defaults to class name).
        """
        return text_type(cls.__name__)

    @classmethod
    def __registered__(cls, reg):
        if not cls.relations:
            raise ValueError(
                '{} is missing a "relations" class attribute'.format(cls))
        # Compute relation/role pairs from relations
        relations = []
        for relinfo in cls.relations:
            try:
                rtype, role = relinfo
            except ValueError:
                rtype = relinfo
                role = 'subject'
            relations.append((rtype, role))
        cls._relations = tuple(relations)
        # Check for name or relation duplication
        for obj in reg[cls.__regid__]:
            if obj == cls:
                continue
            if obj.etype == cls.etype:
                if obj.name == cls.name:
                    # Make sure 'name' is unique amongst objects with
                    # 'jsonschema.object' regid.
                    raise ValueError('a class with name "{}" is already '
                                     'registered'.format(cls.name))
                # Prevent duplicate mapping of the same etype/rtype.
                common_rtypes = set(obj._relations) & set(cls._relations)
                if common_rtypes:
                    raise ValueError(
                        'duplicate relation mapping for "{}": {}'.format(
                            cls.etype,
                            ', '.join('-'.join(rel) for rel in common_rtypes),
                        )
                    )
        # Hide relations mapped to this document from etype JSON Schema.
        for rtype, role in cls._relations:
            if role == 'object':
                jsonschema_section.tag_object_of(
                    ('*', rtype, cls.etype), 'hidden')
            else:
                jsonschema_section.tag_subject_of(
                    (cls.etype, rtype, '*'), 'hidden')
        return super(CompoundMapper, cls).__registered__(reg)

    @property
    def jsl_field_class(self):
        return partial(jsl.fields.DocumentField, as_ref=True,
                       title=self._cw._(self.title))

    def jsl_field(self, schema_role):
        attrs = {}
        for rtype, role in self._relations:
            rschema = self._cw.vreg.schema[rtype]
            target_types = {
                t.type for t in rschema.targets(self.etype, role)}
            mapper = self._cw.vreg['mappers'].select(
                'jsonschema.relation', self._cw,
                etype=self.etype, rtype=rtype, role=role,
                target_types=target_types,
            )
            attrs[mapper.orm_rtype] = mapper.jsl_field(schema_role)
        attrs['Options'] = default_options(self._cw, self.title)
        doc = type(str(self.title), (jsl.Document, ), attrs)
        return self.jsl_field_class(doc)

    def values(self, entity, instance):
        assert entity is None or entity.cw_etype == self.etype, \
            'cannot get "values" for {} with {}'.format(entity, self)
        try:
            values = instance.pop(self.name)
        except KeyError:
            if entity is None:
                return {}
            # Continue with an empty "value" that would get filled by compound
            # relation mappers with default or None values.
            values = {}
        for rtype, role in self._relations:
            mapper = self._relation_mapper(rtype, role)
            values.update(
                mapper.values(entity, values))
        return values

    def serialize(self, entity):
        assert entity.cw_etype == self.etype, \
            'cannot serialize {} with {}'.format(entity, self)
        data = {}
        for rtype, role in self._relations:
            mapper = self._relation_mapper(rtype, role)
            value = mapper.serialize(entity)
            if value is not None:
                data[mapper.orm_rtype] = value
        return data

    def _relation_mapper(self, rtype, role='subject'):
        vreg = self._cw.vreg
        rschema = vreg.schema[rtype]
        target_types = {t.type for t in rschema.targets(self.etype, role)}
        return vreg['mappers'].select(
            'jsonschema.relation', self._cw, etype=self.etype,
            rtype=rtype, role=role, target_types=target_types,
        )


class _RelationMapper(BaseRelationMapper):
    """Abstract class for true relation (as opposed to attribute) mapper.
    """
    __abstract__ = True
    __select__ = ~yams_final_rtype()
    jsl_field_class = jsl.fields.ArrayField

    def jsl_field(self, schema_role, **kwargs):
        if 'items' not in kwargs:
            targets = list(self.jsl_field_targets(schema_role))
            cardinalities = set([])
            rschema = self._cw.vreg.schema[self.rtype]
            for target_type in self.target_types:
                rdef = rschema.role_rdef(
                    self.etype, target_type, self.role)
                cardinalities.add(rdef.role_cardinality(self.role))
            if len(targets) > 1:
                kwargs['items'] = jsl.fields.OneOfField(targets)
            else:
                kwargs['items'] = targets[0]
            if schema_role in (CREATION_ROLE, EDITION_ROLE):
                cardinality = cardinalities.pop()
                if cardinalities:
                    raise BadSchemaDefinition(
                        'inconsistent cardinalities within {0} relation '
                        'definitions'.format(self.rtype))
                if cardinality != '*':
                    if 'required' not in kwargs and cardinality in '1+':
                        kwargs['required'] = True
                    if 'min_items' not in kwargs:
                        kwargs['min_items'] = 0 if cardinality in '?' else 1
                    if 'max_items' not in kwargs and cardinality in '?1':
                        kwargs['max_items'] = 1
        return super(_RelationMapper, self).jsl_field(schema_role, **kwargs)

    def jsl_field_targets(self, schema_role):
        """Return an iterator on jsl field objects to put in the 'items' key."""
        raise NotImplementedError()


class InlinedRelationMapper(_RelationMapper):
    """Map relation as 'inlined', i.e. the target of the relation is
    created/edited along with its original entity.
    """
    __select__ = (_RelationMapper.__select__
                  & yams_component_target()
                  & (match_kwargs('etype') | match_kwargs('entity')))

    def jsl_field_targets(self, schema_role):
        for target_type in self.target_types:
            mapper = self._cw.vreg['mappers'].select(
                'jsonschema.entity', self._cw, etype=target_type,
                rtype=self.rtype, role=neg_role(self.role),
                target_types={self.etype},
            )
            target_document = mapper.jsl_document(schema_role)
            yield jsl.fields.DocumentField(target_document, as_ref=True)

    def values(self, entity, instance):
        # Would require knownledge of the target type from "instance",
        # but the generated JSON schema does not expose this yet.
        assert len(self.target_types) == 1, \
            'cannot handle multiple target types yet: {}'.format(
                self.target_types)
        target_type = self.target_types[0]
        try:
            values = instance.pop(self.rtype)
        except KeyError:
            if entity is None:
                return {}
            values = []
        if not isinstance(values, list):
            # TODO: rather jsonschema.validate()
            raise ValidationError(entity,
                                  {self.rtype: _('value should be an array')})
        if entity is not None:
            # if entity already exists, delete entities related through
            # this mapped relation
            for linked_entity in getattr(entity, self.orm_rtype):
                if linked_entity.cw_etype in self.target_types:
                    linked_entity.cw_delete()
        target_mapper = self._cw.vreg['mappers'].select(
            'jsonschema.entity', self._cw, etype=target_type,
            rtype=self.rtype, role=neg_role(self.role),
            target_types={self.etype},
        )
        result = []
        for subinstance in values:
            subvalues = target_mapper.values(subinstance)
            result.append(self._cw.create_entity(target_type, **subvalues))
        return {self.orm_rtype: result}

    def serialize(self, entity):
        rset = entity.related(
            self.rtype, self.role, targettypes=tuple(self.target_types))
        if not rset:
            return None

        def serialize(entity):
            mapper = self._cw.vreg['mappers'].select(
                'jsonschema.entity', self._cw, entity=entity,
                rtype=self.rtype, role=neg_role(self.role),
                target_types={self.etype},
            )
            return mapper.serialize()

        return [serialize(related) for related in rset.entities()]


class ETypeRelationMapper(_RelationMapper):
    """Map relation as 'generic', i.e. the target of the relation may be
    selected in preexisting possible targets.
    """
    __select__ = (_RelationMapper.__select__
                  & ~yams_component_target()
                  & match_kwargs('etype'))

    def relation_targets(self, schema_role):
        entity = self._cw.vreg['etypes'].etype_class(self.etype)(self._cw)
        potential_targets = []
        for target_type in self.target_types:
            potential_targets.extend(entity.unrelated(
                self.rtype, target_type, self.role).entities())
        return potential_targets

    def jsl_field_targets(self, schema_role):
        choices = []
        for target in self.relation_targets(schema_role):
            choices.append((text_type(target.eid), target.dc_title()))
        yield ChoiceStringField(choices, type='string')

    def values(self, entity, instance):
        try:
            values = instance.pop(self.rtype)
        except KeyError:
            return {}
        if entity is not None:
            entity.cw_set(**{self.orm_rtype: None})
        return {self.orm_rtype: [int(x) for x in values]}

    def serialize(self, entity):
        rset = entity.related(
            self.rtype, self.role, targettypes=tuple(self.target_types))
        if not rset:
            return None
        return [text_type(x.eid) for x in rset.entities()]


class EntityRelationMapper(ETypeRelationMapper):
    """Map relation using existing relations target when role is view,
    fallback to ETypeMapper otherwise to display all possible relation
    targets.
    """
    __select__ = (_RelationMapper.__select__
                  & ~yams_component_target()
                  & match_kwargs('entity'))

    def relation_targets(self, schema_role):
        entity = self.cw_extra_kwargs['entity']
        if schema_role == VIEW_ROLE:
            return entity.related(
                self.rtype, self.role,
                targettypes=tuple(self.target_types)).entities()
        if schema_role == CREATION_ROLE:
            assert len(self.target_types) == 1, \
                'cannot handle multiple target types in {} for {}'.format(
                    self, schema_role)
            targettype = self.target_types[0]
            return entity.unrelated(
                self.rtype, targettype, role=self.role,
            ).entities()
        return super(EntityRelationMapper, self).relation_targets(schema_role)



# pylint: disable=W0622
"""cubicweb-jsonschema application packaging information"""


modname = 'cubicweb_jsonschema'
distname = 'cubicweb-jsonschema'

numversion = (0, 2, 2)
version = '.'.join(str(num) for num in numversion)

license = 'LGPL'
author = 'LOGILAB S.A. (Paris, FRANCE)'
author_email = 'contact@logilab.fr'
description = 'JSON Schema for CubicWeb'
web = 'http://www.cubicweb.org/project/%s' % distname

__depends__ = {
    'cubicweb[pyramid]': '>= 3.24.1',
    'six': '>= 1.4.0',
    'iso8601': None,
    'jsl': None,
    'pyramid': '>= 1.8.0',
}
__recommends__ = {}

classifiers = [
    'Environment :: Web Environment',
    'Framework :: CubicWeb',
    'Programming Language :: Python',
    'Programming Language :: JavaScript',
]



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema hyper schema "link" components.

See http://json-schema.org/draft-06/links for the JSON schema of these
objects.
"""


from cubicweb.appobject import AppObject
from cubicweb.predicates import has_permission, match_kwargs

from .. import VIEW_ROLE, CREATION_ROLE, EDITION_ROLE


class Link(AppObject):
    """Abstract hyper schema link."""
    __registry__ = 'links'

    def description_object(self, request, resource):
        """Return the link description object as a dict."""
        raise NotImplementedError()


class CollectionLink(Link):
    """Link for a collection."""
    __regid__ = 'collection'
    __select__ = match_kwargs('title')
    rel = 'collection'

    @property
    def title(self):
        return self.cw_extra_kwargs['title']

    def description_object(self, request, resource):
        return {
            u'href': request.resource_path(resource),
            u'rel': self.rel,
            u'targetSchema': {
                u'$ref': request.resource_path(resource, 'schema'),
            },
            u'submissionSchema': {
                u'$ref': request.resource_path(resource, 'schema',
                                               query={'role': CREATION_ROLE}),
            },
            u'title': self.title,
        }


class CollectionItemLink(Link):
    """Link for a collection item."""
    __regid__ = 'collection-item'
    __select__ = match_kwargs('anchor')
    rel = 'item'

    def description_object(self, request, resource):
        href = request.resource_path(resource) + '{id}'
        return {
            u'href': href,
            u'rel': self.rel,
            u'anchor': self.cw_extra_kwargs['anchor'],
        }


class ETypeLink(CollectionLink):
    """Link for a collection of entities of the same type."""
    __regid__ = 'collection.etype'
    __select__ = match_kwargs('etype')
    rel = 'self'

    @property
    def title(self):
        etype = self.cw_extra_kwargs['etype']
        return self._cw.__(etype + '_plural')


class EntityLink(Link):
    """Abstract hyper schema link suitable for entity JSON schema."""
    __abstract__ = True
    __select__ = match_kwargs('entity')

    @property
    def entity(self):
        return self.cw_extra_kwargs['entity']


class EntitySelfLink(EntityLink):
    """Link for an entity as rel="self"."""
    __regid__ = 'entity.self'
    __select__ = EntityLink.__select__ & has_permission('read')

    def description_object(self, request, resource):
        entity = self.entity
        title = self._cw._('{0} #{1}').format(
            self._cw.__(entity.cw_etype), entity.eid)
        return {
            u'href': request.resource_path(resource),
            u'rel': u'self',
            u'title': title,
            u'submissionSchema': {
                u'$ref': request.resource_path(
                    resource, 'schema', query={'role': EDITION_ROLE}),
            },
            u'targetSchema': {
                u'$ref': request.resource_path(
                    resource, 'schema', query={'role': VIEW_ROLE}),
            },
        }


class EntityParentLink(EntityLink):
    """Link to the parent resource of an entity (i.e. the entity type)."""
    __regid__ = 'entity.type'
    rel = u'up'

    def description_object(self, request, resource):
        return {
            u'href': request.resource_path(resource.__parent__),
            u'rel': self.rel,
            u'title': self._cw.__('{0}_plural').format(self.entity.cw_etype),
            u'targetSchema': {
                u'$ref': request.resource_path(
                    resource.__parent__, 'schema'),
            },
        }


class EntityCollectionLink(EntityParentLink):
    """Link for an entity as item of a collection."""
    __regid__ = 'entity.collection'
    rel = u'collection'


class EntityRelatedLink(EntityLink):
    """Link to (`rtype`, `role`) relation from `entity` as `role`.
    """
    __regid__ = 'entity.related'
    __select__ = match_kwargs('entity', 'rtype', 'role')

    def description_object(self, request, resource):
        assert resource.__class__.__name__ == 'EntityResource'  # XXX debug
        rtype = self.cw_extra_kwargs['rtype']
        role = self.cw_extra_kwargs['role']
        collection_mapper = self._cw.vreg['mappers'].select(
            'jsonschema.collection', self._cw, rtype=rtype, role=role)
        title = collection_mapper.title
        return {
            u'href': request.resource_path(resource[rtype]),
            u'rel': u'related',
            u'title': title,
        }


class RelatedCollectionLink(CollectionLink):
    """rel="self" link for collection of related entities."""
    __regid__ = 'related-collection'
    rel = 'self'

    def description_object(self, request, resource):
        ldo = super(RelatedCollectionLink, self).description_object(
            request, resource)
        ldo['targetSchema']['$ref'] = request.resource_path(
            resource, 'schema', query={'role': VIEW_ROLE})
        return ldo



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Pyramid views for the JSON API."""

import datetime
from functools import wraps
import logging

from cubicweb import NoResultError

from pyramid import httpexceptions
from pyramid.renderers import JSON, render
from pyramid.view import view_config

from .. import JSONSCHEMA_MEDIA_TYPE


LOG = logging.getLogger(__name__)


def json_config(**settings):
    """Wraps view_config for JSON rendering."""
    settings.setdefault('accept', 'application/json')
    settings.setdefault('renderer', 'json')
    return view_config(**settings)


def link_header(value, rel, **params):
    base = '<{value}>; rel="{rel}"'.format(value=value, rel=rel,)
    extra = '; '.join('{}="{}"'.format(name, value)
                      for name, value in params.items())
    return '; '.join([base, extra])


def describedby(func):
    """View decorator adding a rel="describedby" Link header pointing to
    resource's JSON Schema URI.
    """
    @wraps(func)
    def wrapper(context, request):
        response = func(context, request)
        response.headers['Link'] = link_header(
            request.resource_path(context, 'schema'),
            'describedby', type=JSONSCHEMA_MEDIA_TYPE,
        )
        return response

    return wrapper


def describes(func):
    """View decorator adding a rel="describes" Link header pointing to
    resource's JSON view URI.
    """
    @wraps(func)
    def wrapper(context, request):
        response = func(context, request)
        response.headers['Link'] = link_header(
            request.resource_path(context),
            'describes', type='application/json'
        )
        return response

    return wrapper


def entity_from_context(func):
    """View decorator binding a CubicWeb `entity` to the `context`.

    May raise HTTPNotFound if no entity can be found.
    Will walk through parents of the `context` until an `EntityResource` gets
    found before fetching the `entity`.
    """
    @wraps(func)
    def wrapper(context, request):
        entity_context = context
        while True:
            try:
                rset = entity_context.rset
            except AttributeError:
                try:
                    entity_context = entity_context.__parent__
                except AttributeError:
                    raise httpexceptions.HTTPNotFound()
            else:
                break
        try:
            entity = rset.one()  # May raise HTTPNotFound.
        except NoResultError:
            raise httpexceptions.HTTPNotFound()
        context.entity = entity
        return func(context, request)

    return wrapper


def jsonschema_adapter(cnx, **context):
    return cnx.vreg['adapters'].select('IJSONSchema', cnx, **context)


def json_problem(type=None, title=None, status=400, detail=None,
                 instance=None, **problem):
    """Return a response application/problem+json following RFC 7807"""
    Error = httpexceptions.status_map[status]
    if type in (None, 'about:blank') and title is None:
        # When type 'about:blank' (default), title SHOULD be the same as the
        # recommended HTTP status phrase for that code, although it MAY be
        # localized
        title = Error.title
    for key, value in (
        ('type', type),
        ('title', title),
        ('status', status),
        ('detail', detail),
        ('instance', instance),
    ):
        if value is not None:
            problem[key] = value
    return Error(
        body=render('json', problem),
        headers={'Content-Type': 'application/problem+json; charset=UTF-8'})


def includeme(config):
    json_renderer = JSON()

    def datetime_isoformat(obj, request):
        return obj.replace(microsecond=0).isoformat()

    def date_isoformat(obj, request):
        return obj.isoformat()

    json_renderer.add_adapter(datetime.datetime, datetime_isoformat)
    json_renderer.add_adapter(datetime.date, date_isoformat)
    config.add_renderer('json', json_renderer)
    config.include('.entities')
    config.include('.schema')
    config.scan(__name__)



"""cubicweb-jsonschema application package

JSON Schema for CubicWeb
"""

VIEW_ROLE = 'view'
CREATION_ROLE = 'creation'
EDITION_ROLE = 'edition'

JSONSCHEMA_MEDIA_TYPE = 'application/schema+json'


def orm_rtype(rtype, role):
    if role == 'object':
        return 'reverse_' + rtype
    return rtype



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema unit tests for Pyramid JSON Schema API."""

from __future__ import print_function

import doctest
from functools import partial
from glob import glob
import json
from os import path
import subprocess
import tempfile
import warnings

import jsonschema
from six import PY2

from cubicweb.pyramid.test import PyramidCWTest
from cubicweb_jsonschema import testing


def testcase_cls(test):
    """Return a unittest TestCase subclass to run `test` doctest."""
    # build class name from doctest's filename (slugify, camelcase)
    basename, __ = path.splitext(path.basename(test.filename))
    clsname = ''.join([w.capitalize() for w in basename.split('-')])

    class ApiBaseTC(PyramidCWTest):
        """Base class of API (doc)tests."""

        settings = {
            'cubicweb.bwcompat': False,
            'cubicweb.auth.authtkt.session.secret': 'x',
            'cubicweb.auth.authtkt.persistent.secret': 'x',
            'cubicweb.session.secret': 'x',
        }

        def includeme(self, config):
            config.include('cubicweb_jsonschema.api')

        def logout(self):
            # XXX /logout route is not implemented without bwcompat so reset
            # cookies by hand...
            self.webapp.reset()

        def setup_database(self):
            with self.admin_access.cnx() as cnx:
                for topic_name in (u'fishing', u'gardening', u'sword fish'):
                    cnx.create_entity('Topic', name=topic_name)
                cnx.commit()

    return type(clsname, (ApiBaseTC, ), {})


def load_tests(loader, tests, ignore):

    here = path.dirname(__file__)
    hyperschema = path.join(here, 'data', 'hyper-schema-draft-06.json')
    validationschema = path.join(here, 'data', 'schema-draft-06.json')
    try:
        has_ajv = subprocess.call(['ajv', 'help'], stdout=subprocess.PIPE) == 0
    except OSError:
        warnings.warn('"ajv" is not available, no validation of JSON Schema '
                      'will be performed')
        has_ajv = False

    if PY2:
        temporary_file = tempfile.NamedTemporaryFile
    else:
        temporary_file = partial(tempfile.NamedTemporaryFile,
                                 mode='w', encoding='utf-8')

    def validate_schema(schema):
        """Validate `schema` against validation and hyper meta-schema.

        Rely on ajv validator if available and use python-jsonschema only for
        hyper schema validation since draft-06 is not supported and $ref are
        not apparently resolved (meaning that hyper-schema reference to
        validation schema is not used).
        """
        with open(hyperschema) as f:
            try:
                jsonschema.validate(schema, json.load(f))
            except jsonschema.RefResolutionError as exc:
                warnings.warn(
                    'validation with python-jsonschema failed ({})'.format(
                        str(exc)),
                    RuntimeWarning,
                )
        if not has_ajv:
            return
        with temporary_file(suffix='.json') as schemafile:
            json.dump(schema, schemafile, indent=2)
            schemafile.seek(0)
            for metaschema in (validationschema, hyperschema):
                cmd = ['ajv', 'validate', '-s', metaschema, '-d',
                       schemafile.name]
                proc = subprocess.Popen(
                    cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                stdout, stderr = proc.communicate()
                if proc.returncode != 0:
                    raise AssertionError(
                        '\n'.join([
                            'error validating instance:',
                            json.dumps(schema, indent=2),
                            'against: {}'.format(metaschema),
                            stderr.decode(errors='replace'),
                        ])
                    )

    def get_schema(app, url, status=200, **kwargs):
        """Helper function to get a JSON schema from `url`."""
        response = app.get(url, status=status,
                           headers={'accept': 'application/schema+json'},
                           **kwargs)
        if status != 200:
            return None
        schema = response.json
        validate_schema(schema)
        return response

    def setUp(test):
        tccls = testcase_cls(test)
        testing.setUp(test, tccls, ['webapp', 'login'])
        client = test.globs['client'] = test.globs.pop('webapp')
        client.get_schema = partial(get_schema, client)
        client.login = partial(test.globs.pop('login'))

    def docfilesuite(fpath):
        return doctest.DocFileSuite(
            fpath,
            setUp=setUp,
            tearDown=testing.tearDown,
            module_relative=False,
            checker=testing.WebTestReponseOutputChecker(),
            optionflags=doctest.REPORT_ONLY_FIRST_FAILURE,
        )

    for fpath in glob(path.join(here, 'schema-api', '*.rst')):
        tests.addTests(docfilesuite(fpath))
    for fpath in glob(path.join(here, 'api', '*.rst')):
        tests.addTests(docfilesuite(fpath))
    tests.addTests(
        docfilesuite(path.join(here, 'hypermedia-walkthrough.rst')))

    return tests


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Yams to jsl  unit tests."""

from six import text_type

from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_jsonschema.views.y2jsl import y2j_etype


class CWYams2JsonTC(CubicWebTC):

    def rtype(self, etype, rtype):
        return self.vreg.schema[etype].rdef(rtype)

    def test_auto_attributes(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class CWUser(object):
                pass

        j_schema = CWUser.get_schema(ordered=True)
        j_props = j_schema['properties']
        # Check attribute list
        self.assertEqual(['__etype__', '__eid__', 'login', 'upassword',
                          'firstname', 'surname', 'last_login_time'],
                         list(j_props))
        self.assertEqual(['__etype__', '__eid__', 'login', 'upassword'],
                         j_schema['required'])
        # Check type and description
        login = self.rtype('CWUser', 'login')
        self.assertEqual(login.description,
                         j_props['login'].get('description'))
        self.assertEqual('string', j_props['login'].get('type'))
        # Check format
        self.assertEqual('string', j_props['last_login_time'].get('type'))
        self.assertEqual('date-time', j_props['last_login_time'].get('format'))
        # Check simple constraint
        firstname = self.rtype('CWUser', 'firstname')
        self.assertEqual(firstname.constraint_by_type('SizeConstraint').max,
                         j_props['firstname'].get('maxLength'))

    def test_auto_relations(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class Workflow(object):
                __filter_in__ = ('name', 'description')

            @y2j_etype(self.vreg, cnx)
            class State(object):
                pass

            userworkflow_eid = cnx.execute(
                'Any X WHERE X workflow_of ET, ET name "CWUser"')[0][0]

        j_schema = State.get_schema(ordered=True)
        state_props = j_schema['properties']

        self.assertEqual(['__etype__', '__eid__', 'name', 'description',
                          'state_of'], list(state_props))
        expected_items = {
            'type': 'string',
            'oneOf': [{'enum': [text_type(userworkflow_eid)],
                       'title': u'default user workflow'}],
        }
        self.assertEqual(state_props['state_of']['items'],
                         expected_items)
        self.assertEqual(state_props['state_of']['minItems'], 1)
        self.assertEqual(state_props['state_of']['maxItems'], 1)

    def test_reverse_relation(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class CWUser(object):
                pass

            @y2j_etype(self.vreg, cnx)
            class CWGroup(object):
                __filter_in__ = ('reverse_in_group', )

            users = [(user.eid, user.dc_title())
                     for user in cnx.find('CWUser').entities()]

        j_schema = CWGroup.get_schema(ordered=True)
        properties = j_schema['properties']
        self.assertIn('in_group', properties)
        expected_items = {
            'type': 'string',
            'oneOf': [
                {'enum': [text_type(eid)], 'title': title}
                for eid, title in users],
        }
        self.assertEqual(properties['in_group']['items'],
                         expected_items)

    def test_custom_attribute(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class CWUser(object):
                upassword = {'min_length': 7}

        j_schema = CWUser.get_schema(ordered=True)
        j_props = j_schema['properties']
        self.assertIn('minLength', j_props['upassword'])
        self.assertEqual(7, j_props['upassword']['minLength'])

    def test_custom_relation(self):
        with self.admin_access.cnx() as cnx:

            @y2j_etype(self.vreg, cnx)
            class CWGroup(object):
                pass

            @y2j_etype(self.vreg, cnx)
            class CWUser(object):
                in_group = {'min_items': 1}

        j_schema = CWUser.get_schema(ordered=True)
        user_props = j_schema['properties']
        self.assertIn('in_group', user_props)
        self.assertEqual(1, user_props['in_group']['minItems'])


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema views tests"""


from cubicweb.devtools.testlib import CubicWebTC

from cubicweb_jsonschema.views import (
    jsonschema_section,
)


class JSONSchemaRelationTagsTC(CubicWebTC):

    def test__init(self):
        tagdefs = jsonschema_section._tagdefs
        for tagged, expected_section in (
            # meta rtype
            (('Book', 'eid', 'Int', 'subject'), 'hidden'),
            # ignored rtypes
            (('Book', 'created_by', 'CWUser', 'subject'), 'hidden'),
            # attribute
            (('Book', 'title', 'String', 'subject'), 'inlined'),
            # inlined relation, role-cardinality '1' or '+'
            (('CWUser', 'in_group', 'CWGroup', 'subject'), 'inlined'),
            # inlined relation, composite
            (('CWUser', 'use_email', 'EmailAddress', 'object'), 'inlined'),
            # plain relation
            (('Book', 'topics', 'Topic', 'subject'), 'related'),
            (('Book', 'topics', 'Topic', 'object'), 'related'),
            # computed relation
            (('Author', 'publications', 'Book', 'object'), 'related'),
        ):
            with self.subTest(tagged=tagged):
                self.assertEqual(tagdefs[tagged], expected_section)

    def fake_entity(self, cnx, etype):
        return self.vreg['etypes'].etype_class(etype)(cnx)

    def rtag(self, cnx, entity):
        return self.vreg['uicfg'].select('jsonschema', cnx, entity=entity)

    def test_relations_by_section_password(self):
        """upassword for CWUser is listed for "add" action, not for "read"."""
        with self.admin_access.cnx() as cnx:
            entity = self.fake_entity(cnx, 'CWUser')
            rsection = self.rtag(cnx, entity)
            for rtype, _, _ in rsection.relations_by_section(
                    entity, 'inlined', 'read'):
                if rtype == u'upassword':
                    self.fail('"upassword" unexpectedly found')
            add_rtypes = [
                rtype for rtype, _, _ in rsection.relations_by_section(
                    entity, 'inlined', 'add')]
            self.assertIn(u'upassword', add_rtypes)

    def test_relations_by_section_ignored(self):
        """meta rtypes are not listed."""
        with self.admin_access.cnx() as cnx:
            entity = self.fake_entity(cnx, 'Book')
            rsection = self.rtag(cnx, entity)
            read_rtypes = [
                rtype for rtype, _, _ in rsection.relations_by_section(
                    entity, 'inlined', 'read')]
            self.assertNotIn('eid', read_rtypes)
            add_rtypes = [
                rtype for rtype, _, _ in rsection.relations_by_section(
                    entity, 'related', 'add')]
            self.assertNotIn('custom_workflow', add_rtypes)

    def test_relations_by_section_inlined(self):
        with self.admin_access.cnx() as cnx:
            entity = self.fake_entity(cnx, 'Book')
            rsection = self.rtag(cnx, entity)
            rtypes = [
                rtype for rtype, _, _ in rsection.relations_by_section(
                    entity, 'inlined', 'read')]
            self.assertCountEqual(['title', 'publication_date'], rtypes)

    def test_relations_by_section_related(self):
        with self.admin_access.cnx() as cnx:
            entity = self.fake_entity(cnx, 'Book')
            rsection = self.rtag(cnx, entity)
            rtypes = [rtype for rtype, _, _ in rsection.relations_by_section(
                entity, 'related', 'read')]
            self.assertCountEqual(['author', 'publications', 'topics'],
                                  rtypes)

    def test_relations_by_section_inlined_permissions(self):
        with self.admin_access.cnx() as cnx:
            cnx.create_entity('Book', title=u't')
            self.create_user(cnx, u'bob')
            cnx.commit()
        with self.new_access(u'bob').cnx() as cnx:
            entity = cnx.find('Book').one()
            rsection = self.rtag(cnx, entity)
            read_rtypes = [
                rtype for rtype, _, _ in rsection.relations_by_section(
                    entity, 'inlined', 'read')]
            self.assertCountEqual(['title', 'publication_date'], read_rtypes)
            update_rtypes = [
                rtype for rtype, _, _ in rsection.relations_by_section(
                    entity, 'inlined', 'update')]
            self.assertCountEqual([], update_rtypes)

    def test_relations_by_section_related_permissions(self):
        with self.admin_access.cnx() as cnx:
            author = cnx.create_entity('Author', name=u'a')
            cnx.create_entity('Book', title=u't', author=author)
            self.create_user(cnx, u'bob')
            cnx.commit()
        with self.new_access(u'bob').cnx() as cnx:
            entity = cnx.find('Book').one()
            rsection = self.rtag(cnx, entity)
            update_rtypes = [
                rtype for rtype, _, _ in rsection.relations_by_section(
                    entity, 'related', 'add')]
            self.assertCountEqual(['topics'],
                                  update_rtypes)


if __name__ == '__main__':
    import unittest
    unittest.main()



from cubicweb.web.views import uicfg

from cubicweb_jsonschema.views import jsonschema_section



uicfg.indexview_etype_section['Photo'] = 'subobject'
uicfg.indexview_etype_section['Topic'] = 'subobject'

jsonschema_section.tag_attribute(
    ('CWUser', 'lastname'), 'hidden')
jsonschema_section.tag_subject_of(
    ('CWUser', 'use_email', 'EmailAddress'), 'inlined')
jsonschema_section.tag_subject_of(
    ('CWUser', 'use_email', 'EmailAlias'), 'hidden')
jsonschema_section.tag_object_of(
    ('*', 'use_email', 'EmailAlias'), 'hidden')
jsonschema_section.tag_subject_of(
    ('CWUser', 'picture', '*'), 'inlined')
jsonschema_section.tag_object_of(
    ('*', 'picture', 'Photo'), 'hidden')
jsonschema_section.tag_subject_of(
    ('Photo', 'thumbnail', '*'), 'inlined')

# we don't want this to interfere with entity creation/edition tests
jsonschema_section.tag_subject_of(
    ('CWUser', 'use_email', 'EmailAlias'), 'hidden')



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema unit tests for Pyramid JSON API."""

import base64
from datetime import date
from unittest import skip

import jsonschema
from mock import patch

from cubicweb import Binary, ValidationError
from cubicweb.pyramid.test import PyramidCWTest

from cubicweb_jsonschema import VIEW_ROLE
from cubicweb_jsonschema.entities.ijsonschema import (
    jsonschema_adapter,
)


class BaseTC(PyramidCWTest):

    settings = {
        'cubicweb.bwcompat': False,
        'pyramid.debug_notfound': True,
        'cubicweb.auth.authtkt.session.secret': 'x',
        'cubicweb.auth.authtkt.persistent.secret': 'x',
        'cubicweb.session.secret': 'x',
    }

    def includeme(self, config):
        config.include('cubicweb_jsonschema.api')


class EntitiesTC(BaseTC):

    @skip('todo')
    def test_post_json_file_upload(self):
        data = {
            'login': u'bob',
            'upassword': u'bob',
            'picture': [{
                u'data': u'data:text/xml;name=test;base64,{}'.format(
                    base64.b64encode('hello')),
            }],
        }
        self.login()
        resp = self.webapp.post_json('/cwuser/', data, status=201,
                                     headers={'Accept': 'application/json'})
        with self.admin_access.cnx() as cnx:
            entity = cnx.find('CWUser', login=u'bob').one()
            self.assertTrue(entity.picture)
            photo = entity.picture[0]
            self.assertEqual(photo.read(), 'hello')
        self.assertEqual(
            resp.location, 'https://localhost:80/CWUser/%d' % entity.eid)

    @skip('todo')
    def test_post_json_file_upload_badrequest(self):
        self.login()
        for rtype, value in [
            ('unknown', [{'data': u'who cares?'}]),
            ('picture', [{'data': u'badprefix:blah blah'}]),
            ('picture', {'data': u'not in a list'}),
        ]:
            data = {rtype: value}
            with self.subTest(**data):
                data[u'login'] = u'bob'
                data[u'upassword'] = u'bob'
                # Rely on "status=400" for test assertion.
                self.webapp.post_json('/CWUser/', data, status=400,
                                      headers={'Accept': 'application/json'})

    def test_get_related(self):
        with self.admin_access.cnx() as cnx:
            book = cnx.create_entity('Book', title=u'title')
            author = cnx.create_entity('Author', name=u'bob',
                                       reverse_author=book)
            jschema = self.vreg['adapters'].select(
                'IJSONSchema', cnx, entity=author,
                rtype='author', role='object').view_schema()
            cnx.commit()
        url = '/book/%s/author' % book.eid
        res = self.webapp.get(url, headers={'accept': 'application/json'})
        related = res.json
        expected = [{
            'type': u'author',
            'id': str(author.eid),
            'title': u'bob',
        }]
        with self.admin_access.cnx() as cnx:
            collection_mapper = cnx.vreg['mappers'].select(
                'jsonschema.collection', cnx, etype='Author')
            jschema = collection_mapper.jsl_field(VIEW_ROLE).get_schema()
        jsonschema.validate(related, jschema)
        self.assertEqual(related, expected)

    def test_post_related_bad_identifier(self):
        with self.admin_access.cnx() as cnx:
            book = cnx.create_entity('Book', title=u'title')
            cnx.commit()
        self.login()
        url = '/book/%s/author' % book.eid
        res = self.webapp.post_json(url, ['a'], status=400,
                                    headers={'accept': 'application/json'})
        self.assertIn('invalid target identifier(s)', res.json['message'])

    def test_post_related_bad_role(self):
        with self.admin_access.cnx() as cnx:
            author = cnx.create_entity('Author', name=u'bob')
            cnx.commit()
        self.login()
        url = '/author/%s/author' % author.eid
        res = self.webapp.post_json(url, [], status=404,
                                    headers={'accept': 'application/json'})
        self.assertIn('relation author-subject not found on Author',
                      res.json['message'])

    def test_post_related_entity_notfound(self):
        self.login()
        url = '/book/999/author'
        self.webapp.post_json(url, [], status=404,
                              headers={'accept': 'application/json'})

    def test_post_related_bad_target(self):
        with self.admin_access.cnx() as cnx:
            book = cnx.create_entity('Book', title=u'title')
            cnx.commit()
            cwuser_eid = cnx.find('CWUser', login=u'admin')[0][0]
        self.login()
        url = '/book/%s/author' % book.eid
        # TODO: status should be 4xx
        self.webapp.post_json(url, [str(cwuser_eid)], status=500,
                              headers={'accept': 'application/json'})

    def test_put_with_incomplete_data(self):
        """A PUT request *replaces* entity attributes, so if fields are
        missing from JSON request body, respective attributes are reset.
        """
        with self.admin_access.cnx() as cnx:
            entity = cnx.create_entity('Photo', data=Binary(b'plop'),
                                       flash=True,
                                       exposure_time=1.23)
            cnx.create_entity('Thumbnail', data=Binary(b'plip'),
                              reverse_thumbnail=entity)
            cnx.commit()
        self.login()
        url = '/photo/{}/'.format(entity.eid)
        self.webapp.put_json(url, {'data': 'plip', 'media_type': u'jpeg'},
                             headers={'Accept': 'application/json'})
        with self.admin_access.cnx() as cnx:
            entity = cnx.entity_from_eid(entity.eid)
            self.assertEqual(entity.data.getvalue(), b'plip')
            self.assertEqual(entity.media_type, u'jpeg')
            # 'thumbnail' absent from request body, we should get ().
            self.assertEqual(entity.thumbnail, ())
            self.assertFalse(cnx.find('Thumbnail'))
            # 'flash' has a default value, we should get this back.
            self.assertEqual(entity.flash, False)
            # 'exposure_time' absent from request body, we should get None.
            self.assertIsNone(entity.exposure_time)

    def test_get_related_sort(self):
        """Sort by modification_date ascending and descending"""
        with self.admin_access.cnx() as cnx:
            author = cnx.create_entity(
                'Author', name=u'bob')
            book1 = cnx.create_entity(
                'Book', title=u'1',
                publication_date=date(1976, 3, 1),
                author=author,
            )
            book2 = cnx.create_entity(
                'Book', title=u'1',
                publication_date=date(1977, 3, 1),
                author=author,
            )
            cnx.commit()

        ascending = [book1.title, book2.title]
        descending = ascending[::-1]
        for sort, expected in [
            ('publication_date', ascending),
            ('-publication_date', descending),
        ]:
            with self.subTest(sort=sort):
                url = ('/author/%s/publications?sort=%s'
                       % (author.eid, sort))
                res = self.webapp.get(
                    url, headers={'accept': 'application/json'})
                entities = res.json
                self.assertEqual(len(entities), 2)
                ids = [d['title'] for d in entities]
                self.assertEqual(ids, expected)

    def test_add_related(self):
        """POST on /<etype>/<eid>/relationships/<rtype> with primary entity as
        subject of <rtype>.
        """
        with self.admin_access.repo_cnx() as cnx:
            book = cnx.create_entity('Book', title=u'tmp')
            cnx.commit()
        url = '/book/%d/relationships/author' % book.eid
        data = {
            'name': u'bob',
        }
        self.login()
        res = self.webapp.post_json(url, data, status=201,
                                    headers={'accept': 'application/json'})
        entity = res.json
        with self.admin_access.cnx() as cnx:
            author_eid = cnx.find('Author').one().eid
            jschema = jsonschema_adapter(
                cnx, etype='Author').view_schema()
        self.assertEqual(
            res.location,
            'https://localhost:80/author/{}/'.format(author_eid))
        jsonschema.validate(entity, jschema)
        self.assertEqual(entity['name'], u'bob')
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute(
                'Any A WHERE B author A, B eid %(b)s, A name "bob"',
                {'b': book.eid})
        self.assertTrue(rset)

    def test_add_related_object(self):
        """POST on /<etype>/<eid>/relationships/<rtype> with primary entity as
        object of <rtype>.
        """
        with self.admin_access.repo_cnx() as cnx:
            author = cnx.create_entity('Author', name=u'bob')
            cnx.commit()
        url = '/author/%s/relationships/author' % author.eid
        data = {
            'title': u'introducing cubicweb-jsonschema',
        }
        self.login()
        res = self.webapp.post_json(url, data, status=201,
                                    headers={'accept': 'application/json'})
        entity = res.json
        with self.admin_access.cnx() as cnx:
            book_eid = cnx.find('Book').one().eid
            jschema = jsonschema_adapter(
                cnx, etype='Book').view_schema()
        self.assertEqual(
            res.location,
            'https://localhost:80/book/{}/'.format(book_eid))
        jsonschema.validate(entity, jschema)
        self.assertEqual(entity['title'], u'introducing cubicweb-jsonschema')
        with self.admin_access.cnx() as cnx:
            rset = cnx.execute(
                'Any X WHERE X author A, A eid %(a)s,'
                ' X title "introducing cubicweb-jsonschema"',
                {'a': author.eid})
        self.assertTrue(rset)

    def test_validationerror_additionalproperties(self):
        data = {
            u'name': u'bob',
            u'born': u'1986',
        }
        self.login()
        res = self.webapp.post_json('/author/', data, status=422,
                                    headers={'Accept': 'application/json'})
        self.assertEqual(res.content_type, 'application/problem+json')
        self.assertEqual(res.json_body, {
            u'status': 422,
            u'title': u'Unprocessable Entity',
            u'detail': u'unexpected properties: born',
        })

    def test_validationerror_nosource(self):
        """Test validation_failed view with no specific source entry."""
        with self.admin_access.cnx() as cnx:
            gid = cnx.find('CWGroup', name=u'users')[0][0]
        data = {
            'login': u'bob',
            'upassword': u'pass',
            'in_group': [str(gid)],
        }
        with patch('cubicweb.req.RequestSessionBase.create_entity',
                   side_effect=ValidationError(None, {None: u'unmapped'})):
            self.login()
            res = self.webapp.post_json('/cwuser/', data, status=422,
                                        headers={'Accept': 'application/json'})
            self.assertEqual(res.content_type, 'application/problem+json')
            self.assertEqual(res.json_body, {
                u'status': 422,
                u'title': u'Unprocessable Entity',
                u'detail': u'unmapped',
            })

    def test_validationerror_integrityerror(self):
        with self.admin_access.cnx() as cnx:
            gid = cnx.find('CWGroup', name=u'users')[0][0]
            self.create_user(cnx, u'bob')
            cnx.commit()
        data = {
            'login': 'bob',
            'upassword': 'pass',
            'in_group': [str(gid)],
        }
        self.login()
        res = self.webapp.post_json('/cwuser/', data, status=422,
                                    headers={'Accept': 'application/json'})
        self.assertEqual(res.content_type, 'application/problem+json')
        self.assertEqual(res.json_body, {
            u'status': 422,
            u'title': u'Unprocessable Entity',
            u'detail': u'some relations violate a unicity constraint',
            u'invalid-params': [
                {u'name': u'login',
                 u'reason': u'login is part of violated unicity constraint'}],
        })

    def test_exception(self):
        with self.admin_access.cnx() as cnx:
            gid = cnx.find('CWGroup', name=u'users')[0][0]
        data = {
            'login': 'bob',
            'upassword': 'pass',
            'in_group': [str(gid)],
        }
        self.login()
        error = ValueError("invalid literal for int() with base 10: 'foo'")
        with patch('cubicweb.req.RequestSessionBase.create_entity',
                   side_effect=error):
            res = self.webapp.post_json('/cwuser/', data, status=500,
                                        headers={'Accept': 'application/json'})
            self.assertEqual(res.content_type, 'application/problem+json')
            self.assertEqual(res.json_body, {
                u'status': 500,
                u'title': u'Internal Server Error',
                u'detail': u"invalid literal for int() with base 10: 'foo'",
            })

    def test_exception_no_detail(self):
        data = {'title': 'thing'}
        self.login()

        with patch('cubicweb.req.RequestSessionBase.create_entity',
                   side_effect=NotImplementedError()):
            res = self.webapp.post_json('/book/', data, status=500,
                                        headers={'Accept': 'application/json'})
            self.assertEqual(res.content_type, 'application/problem+json')
            self.assertEqual(res.json_body, {
                u'status': 500,
                u'title': u'Internal Server Error',
            })


if __name__ == '__main__':
    import unittest
    unittest.main()



from base64 import b64encode, b64decode

from cubicweb import _, Binary

from cubicweb_jsonschema.mappers import (
    BytesMapper,
    CompoundMapper,
    StringMapper,
    yams_match,
)


class EMailMapper(StringMapper):
    __select__ = yams_match(etype='EmailAddress',
                            rtype='address', role='subject')
    format = 'email'


class ThumbnailDataMapper(BytesMapper):
    __select__ = yams_match(etype='Thumbnail',
                            rtype='data', role='subject',
                            target_types='Bytes')

    @staticmethod
    def _type(value):
        return Binary(b64decode(value))

    @staticmethod
    def _value(value):
        return b64encode(value.getvalue()).decode('ascii')


class exif_data(CompoundMapper):
    etype = 'Photo'
    relations = ('exposure_time', 'flash', 'maker_note')
    title = _('EXIF data')



"""Tests for "testing" module."""
from __future__ import print_function

from functools import partial
import doctest
import unittest

import webtest

from cubicweb_jsonschema import testing


class WebTestReponseOutputCheckerTC(unittest.TestCase):

    def setUp(self):
        self.checker = testing.WebTestReponseOutputChecker()

    def example(self, want):
        """Return a mock of doctest.Example suitable for output_difference()
        usage.
        """
        return type('Example', (object, ), {'want': want})()

    def test_ok(self):
        got = want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"id": "1", "login": "alice"}',
        ])
        self.assertTrue(self.checker.check_output(want, got, 0))
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0), '')

    def test_ok_json_body_prettyformat(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"id": "1", "login": "alice"}',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{',
            '  "id": "1",',
            '  "login": "alice"',
            '}',
        ])
        self.assertTrue(self.checker.check_output(want, got, 0))
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0), '')

    def test_json_ellipsis_object(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"id": "1", "login": "alice"}',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{'
            '  "id": "...",'
            '  "login": "alice"'
            '}',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))
        self.assertTrue(self.checker.check_output(want, got, doctest.ELLIPSIS))

    def test_json_ellipsis_list(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '[{"id": "1", "login": "alice"}]',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '['
            '  {"id": "...", "login": "alice"}'
            ']',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))
        self.assertTrue(self.checker.check_output(want, got, doctest.ELLIPSIS))

    def test_bad_json(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{I am invalid JSON!',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"id": "1"}',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))
        expected = '\n'.join([
            'Actual body is invalid JSON: {I am invalid JSON!',
        ])
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0),
            expected)

    def test_status_differ(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"id": "1", "login": "alice"}',
        ])
        want = '\n'.join([
            'Response: 404 Not Found',
            'Content-Type: application/json',
            '{',
            '  "id": "1",',
            '  "login": "alice"',
            '}',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))
        expected = '\n'.join([
            'Expected status: Response: 404 Not Found',
            'Actual status: Response: 200 OK',
        ])
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0),
            expected)

    def test_headers_differ(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/octet-stream',
            '{"x": 1}',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"x": 1}',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))
        expected = '\n'.join([
            'Expected headers: Content-Type: application/json',
            'Actual headers: Content-Type: application/octet-stream',
        ])
        self.assertMultiLineEqual(
            self.checker.output_difference(self.example(want), got, 0),
            expected)

    def test_headers_missing_from_got(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"x": 1}',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            'Link: </schema> rel="describedby"',
            '{"x": 1}',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))

    def test_headers_extra_in_got(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            'Something: we do not care about',
            '{"x": 1}',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"x": 1}',
        ])
        self.assertTrue(self.checker.check_output(want, got, 0))

    def test_headers_ellipsis(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Location: http://www.example.com/123',
            'Content-Type: application/json',
            '{"x": 1}',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            'Location: http://www.example.com/...',
            '{"x": 1}',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))
        self.assertTrue(self.checker.check_output(want, got, doctest.ELLIPSIS))
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            'Location: http://www.example.org/...',
            '{"x": 1}',
        ])
        self.assertFalse(self.checker.check_output(
            want, got, doctest.ELLIPSIS))

    def test_headers_ellipsis_body_diff(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Location: http://www.example.com/123',
            'Content-Type: application/json',
            '{"id": "1"}',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            'Location: http://www.example.com/...',
            '{"id": "2"}',
        ])
        self.assertFalse(self.checker.check_output(
            want, got, doctest.ELLIPSIS))
        difference = self.checker.output_difference(self.example(want), got, 0)
        # Do not check full content until output_difference() handles ELLIPSIS
        # flag.
        self.assertIn('JSON bodies differ', difference)

    def test_json_body_differ(self):
        got = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"id": "1"}',
        ])
        want = '\n'.join([
            'Response: 200 OK',
            'Content-Type: application/json',
            '{"id": "2"}',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))
        self.assertFalse(self.checker.check_output(
            want, got, doctest.ELLIPSIS))
        expected = '\n'.join([
            'JSON bodies differ',
            'expected',
            '{',
            '  "id": "2"',
            '}',
            'actual',
            '{',
            '  "id": "1"',
            '}',
            'diff:',
            '{"id": ["2", "1"]}',
        ])
        difference = self.checker.output_difference(self.example(want), got, 0)
        self.assertMultiLineEqual(difference, expected)

    def test_no_content(self):
        got = 'Response: 204 No Content<BLANKLINE><BLANKLINE>'
        want = 'Response: 204 No Content<BLANKLINE>'
        self.assertTrue(self.checker.check_output(want, got, 0))
        got = 'Response: 204 No Content\nFoo: bar\n'
        want = 'Response: 204 No Content\nFoo: bar\n{"x": 1}'
        self.assertFalse(self.checker.check_output(want, got, 0))
        got = 'Response: 204 No Content\nFoo: bar\n{"x": 1}'
        want = 'Response: 204 No Content\nFoo: bar\n<BLANKLINE>'
        self.assertFalse(self.checker.check_output(want, got, 0))

    def test_parse_error(self):
        got = '\n'.join([
            'Response 42',
        ])
        want = '\n'.join([
            'Response: OK',
            'Content-Type: something',
            'bobybody',
        ])
        self.assertFalse(self.checker.check_output(want, got, 0))
        expected = '\n'.join([
            'In expected: not a webtest.TestResponse string', want,
            'In actual: not a webtest.TestResponse string', got,
        ])
        difference = self.checker.output_difference(self.example(want), got, 0)
        self.assertMultiLineEqual(difference, expected)


def unittest_integration_test():
    """Return a DocFileSuite instance using "doctest-example.txt" file with
    a setUp coming from a unittest TestCase.
    """
    class WebTestBaseCase(unittest.TestCase):
        """Base TestCase with a custom setup initializing a webtest
        application.
        """

        def setUp(self):
            def application(environ, start_response):
                body = b'{"a": 1, "b": [1, 2]}'
                headers = [('Content-Type', 'application/json; charset=utf8'),
                           ('Content-Length', str(len(body)))]
                start_response('200 OK', headers)
                return [body]

            self.webapp = webtest.TestApp(application)

    setUp = partial(testing.setUp, cls=WebTestBaseCase,
                    globs_from_testcase=['webapp'])

    return doctest.DocFileSuite(
        'doctest-example.txt', setUp=setUp,
        checker=testing.WebTestReponseOutputChecker())


def load_tests(loader, tests, ignore):
    """unittest discovery callback"""
    tests.addTests(
        unittest_integration_test(),
    )
    tests.addTests(
        doctest.DocTestSuite('cubicweb_jsonschema.testing'),
    )
    return tests


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from collections import OrderedDict

import jsl

from cubicweb.schema import META_RTYPES
from cubicweb.view import Component

from cubicweb_jsonschema import CREATION_ROLE


class Y2JError(Exception):
    pass


def field_names_from_yams(etype):
    """ Yield the names of jsl fields for entity type ``etype``.
    These names are yams attributes and inlined relations.
    """
    for rtype in etype.ordered_relations():
        if rtype.type not in META_RTYPES:
            if etype.is_metadata(rtype.type):
                continue
            if rtype.final or rtype.inlined:
                yield rtype.type, 'subject'


class JslDocumentBuilder(Component):
    """Class that builds a jsl document from a description class.

    Inherit this class and pass an instance to `y2j_etype` to
    customize the jsl field gathering and generation.

    By default, the resulting jsl document will have following fields,
    in this order:

    - `__etype__` and `__eid__`

    - all yams attributes and inlined relations (unless an explicit
      name list is supplied in the `from_yams` attribute of the
      description class)

    - all the jsl fields attributes set on the description class

    The field generation itself is be delegated to the classes which
    `__regid__` is "jsonschema.relation" in the "mappers" register.

    """
    __regid__ = 'jsonschema.map.builder'

    # Should "base" fields be included in main schema document?
    include_base_fields = False

    def __init__(self, *args, **kwargs):
        super(JslDocumentBuilder, self).__init__(*args, **kwargs)
        self.relation_targets = {}

    def base_fields(self, etype):
        fields = OrderedDict()
        fields['__etype__'] = jsl.fields.StringField(
            required=True, enum=[etype], default=etype)
        fields['__eid__'] = jsl.fields.StringField(required=True)
        return fields

    def fields_from_yams(self, etype, filter_in=None, filter_out=None,
                         schema_role=None):
        """Return fields for a Yams `etype` using the `generate_field` method.

        The field list is built using `field_names_from_yams` for yams entity
        type which name is `etype`, unless `filter_in` specifies such a field
        name list.

        If `filter_out` is given, it must be a sequence of names which are
        excluded from the above field name list. Therefore, if a name is both
        in `filter_in` and `filter_out`, an error is raised.
        """
        if filter_in and filter_out:
            both = set(filter_in).intersection(set(filter_out))
            if both:
                msg = self._cw._(
                    'Fields %s are both in __filter_in__ and __filter_out__')
                raise Y2JError(msg % both)
        if filter_in:
            rtype_roles = []
            for value in filter_in:
                try:
                    rtype, role = value
                except ValueError:
                    if value.startswith('reverse_'):
                        rtype, role = value[len('reverse_'):], 'object'
                    else:
                        rtype, role = value, 'subject'
                rtype_roles.append((rtype, role))
        else:
            rtype_roles = field_names_from_yams(self._cw.vreg.schema[etype])
        if filter_out:
            rtype_roles = [rtype_role for rtype_role in rtype_roles
                           if rtype_role[0] not in filter_out]
        return OrderedDict((rtype, self.generate_field(etype, rtype, role,
                                                       schema_role))
                           for rtype, role in rtype_roles)

    def explicit_fields(self, etype, explicit_fields, schema_role=None):
        """Collect attributes from `explicit_fields` mapping.

        Values of `explicit_fields` may either jsl fields or dict instances.
        In the later case, the field is built using a field factory
        which is passed keyword arguments given at the dict value.
        """
        fields = OrderedDict()
        for key, val in explicit_fields.items():
            if isinstance(val, jsl.fields.BaseField):
                fields[key] = val
            elif isinstance(val, dict):
                fields[key] = self.generate_field(
                    etype, key, 'subject', schema_role, **val)
        return fields

    def generate_field(self, etype, rtype, role, schema_role,
                       target_types=None, **kwargs):
        """ Return a jsl field named `rtype` supposed to described a
        yams relation definition of entity type `etype` as `role` in the
        context of schema `schema_role` (in the jsl sense).

        The generation is delegated to a field factory which is called
        with the keyword arguments `kwargs`.
        """
        if target_types is None:
            rschema = self._cw.vreg.schema[rtype]
            target_types = {
                target.type for target in rschema.targets(etype, role)}
        assert isinstance(target_types, set), target_types
        field_factory = self._cw.vreg['mappers'].select_or_none(
            'jsonschema.relation', self._cw,
            etype=etype, rtype=rtype, role=role, target_types=target_types)
        if field_factory is None:
            raise Y2JError(
                'Cannot find a field factory for %s-%s' % (etype, rtype))
        fieldcls = field_factory.jsl_field(schema_role, **kwargs)
        assert (rtype, role) not in self.relation_targets, \
            'field for {0}-{1} already generated'.format(rtype, role)
        self.relation_targets[rtype, role] = field_factory.target_types
        return fieldcls

    def set_default_options(self, etype, options=None):
        """Return an Options class with `definition_id` to the class name of
        `etype`; it `options` is not None and is an `Options` class, it is
        updated and returned.
        """
        if options is None:
            options = type('Options', (object,), {})
        if 'definition_id' not in dir(options):
            options.definition_id = etype
        if 'title' not in dir(options):
            options.title = self._cw._(etype)
        return options

    def document(self, cls, schema_role=None):
        """ Return a Jsl Document class from the description class `cls`.
        This is the entry point of the builder.
        """
        etype = cls.__name__
        newdict = self.base_fields(etype)
        fields = dict((k, getattr(cls, k)) for k in dir(cls))
        # Add explicit fields to `filter_out` to avoid generating them twice.
        filter_out = getattr(cls, '__filter_out__', ()) + tuple(fields)
        yams_fields = self.fields_from_yams(
            etype,
            filter_in=getattr(cls, '__filter_in__', ()),
            filter_out=filter_out,
            schema_role=schema_role,
        )
        newdict.update(yams_fields)
        newdict.update(self.explicit_fields(etype, fields, schema_role))
        options = newdict.get('Options')
        newdict['Options'] = self.set_default_options(etype, options)
        return type(etype, (jsl.Document,), newdict)


def y2j_etype(vreg, cnx, schema_role=CREATION_ROLE):
    """Class decorator that returns a jsl Document subclass.

    This jsl document is built using registered 'jsonschema.map.builder'
    component.
    """
    def decorator(cls):
        builder = vreg['components'].select('jsonschema.map.builder', cnx)
        return builder.document(cls, schema_role)
    return decorator



#!/usr/bin/env python
# pylint: disable=W0142,W0403,W0404,W0613,W0622,W0622,W0704,R0904,C0103,E0611
#
# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of a cubicweb-jsonschema.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""cubicweb_jsonschema setup module using data from
cubicweb_jsonschema/__pkginfo__.py file
"""

from os.path import join, dirname

from setuptools import find_packages, setup


here = dirname(__file__)

# load metadata from the __pkginfo__.py file so there is no risk of conflict
# see https://packaging.python.org/en/latest/single_source_version.html
pkginfo = join(here, 'cubicweb_jsonschema', '__pkginfo__.py')
__pkginfo__ = {}
with open(pkginfo) as f:
    exec(f.read(), __pkginfo__)

# get required metadatas
distname = __pkginfo__['distname']
version = __pkginfo__['version']
license = __pkginfo__['license']
description = __pkginfo__['description']
web = __pkginfo__['web']
author = __pkginfo__['author']
author_email = __pkginfo__['author_email']
classifiers = __pkginfo__['classifiers']

with open(join(here, 'README.rst')) as f:
    long_description = f.read()

# get optional metadatas
data_files = __pkginfo__.get('data_files', None)
dependency_links = __pkginfo__.get('dependency_links', ())

requires = {}
for entry in ("__depends__",):  # "__recommends__"):
    requires.update(__pkginfo__.get(entry, {}))
install_requires = ["{0} {1}".format(d, v and v or "").strip()
                    for d, v in requires.items()]


setup(
    name=distname,
    version=version,
    license=license,
    description=description,
    long_description=long_description,
    author=author,
    author_email=author_email,
    url=web,
    classifiers=classifiers,
    packages=find_packages(exclude=['test']),
    install_requires=install_requires,
    include_package_data=True,
    entry_points={
        'cubicweb.cubes': [
            'jsonschema=cubicweb_jsonschema',
        ],
    },
    zip_safe=False,
)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-jsonschema Pyramid "entities" resources definitions."""

from functools import wraps

from six import text_type

from pyramid.decorator import reify
from pyramid.httpexceptions import HTTPNotFound

from rql import nodes, TypeResolverException


def parent(init_method):
    """Decorator for resource class's __init__ method to bind the __parent__
    attribute to instance.
    """
    @wraps(init_method)
    def wrapper(self, *args, **kwargs):
        parent = kwargs.pop('parent', None)
        init_method(self, *args, **kwargs)
        self.__parent__ = parent

    return wrapper


def find_relationship(schema, etype, rtype):
    """Return (`rtype`, `role`) if `etype` has a relationship `rtype` else
    raise a ValueError.
    """
    eschema = schema[etype]
    for relinfo in eschema.relation_definitions():
        rschema, __, role = relinfo
        if rschema.final or rschema.meta:
            continue
        if rschema.type == rtype:
            return relinfo
    raise ValueError(rtype)


class RootResource(object):
    """Root resources for entities."""
    __parent__ = None
    __name__ = ''

    def __init__(self, request):
        self.request = request

    def __getitem__(self, value):
        try:
            return ETypeResource(self.request, value, self)
        except KeyError:
            raise KeyError(value)


class ETypeResource(object):
    """A resource class for an entity type."""

    @classmethod
    def from_match(cls, matchname):
        def factory(request):
            return cls(request, request.matchdict[matchname])
        return factory

    def __init__(self, request, etype, parent=None):
        vreg = request.registry['cubicweb.registry']
        self.request = request
        self.etype = vreg.case_insensitive_etypes[etype.lower()]
        self.cls = vreg['etypes'].etype_class(self.etype)
        if parent is None:
            parent = RootResource(request)
        self.__parent__ = parent
        self.__name__ = self.etype.lower()

    def __getitem__(self, value):
        if value == 'relationships':
            return _RelationshipProxyResource(self.request, self.etype)
        for attrname in ('eid', self.cls.cw_rest_attr_info()[0]):
            resource = EntityResource(self.request, self.cls,
                                      attrname, value, self)
            try:
                rset = resource.rset
            except HTTPNotFound:
                continue
            if rset.rowcount:
                return resource
        raise KeyError(value)

    @reify
    def rset(self):
        rql = self.cls.fetch_rql(self.request.cw_cnx.user)
        rset = self.request.cw_request.execute(rql)
        return rset


class _RelationshipProxyResource(object):
    """A "proxy" resource allowing to pass through the "relationships" segment
    of path during traversal.
    """

    @parent
    def __init__(self, request, etype=None, **kwargs):
        self.request = request
        self.etype = etype

    def __getitem__(self, value):
        vreg = self.request.registry['cubicweb.registry']
        schema = vreg.schema
        if self.etype:
            etype = self.etype
        elif hasattr(self, '__parent__'):
            etype = self.__parent__.entity.cw_etype
        else:
            raise AssertionError('neither etype nor __parent__ got specified')
        try:
            relinfo = find_relationship(schema, etype, value)
        except ValueError:
            raise KeyError(value)
        else:
            parent = getattr(self, '__parent__', None)
            return RelationshipResource(self.request, *relinfo, parent=parent)


class RelationshipResource(object):
    """A resource to manipulate a relationship `rtype` with __parent__ as
    `role`.

    This can be used for both "schema" and "entities" resources.
    """

    @parent
    def __init__(self, request, rschema, target_schemas, role, **kwargs):
        self.request = request
        self.rtype = rschema.type
        self.target_schemas = target_schemas
        self.role = role

    @property
    def target_type(self):
        """Target entity type of relationship."""
        if len(self.target_schemas) > 1:
            target = self.request.params.get('target_type')
            if target and target in {s.type for s in self.target_schemas}:
                return target
            raise NotImplementedError()
        return self.target_schemas[0].type


class EntityResource(object):
    """A resource class for an entity. It provide method to retrieve an entity
    by eid.
    """

    def __init__(self, request, cls, attrname, name, parent):
        self.request = request
        self.cls = cls
        self.attrname = attrname
        self.__name__ = name
        self.__parent__ = parent

    def __getitem__(self, value):
        # Make sure parent class has no __getitem__ as we do not call it here.
        # If it happens, we'll need to call it here.
        assert not hasattr(super(EntityResource, self), '__getitem__')
        etype = self.cls.cw_etype
        if value == 'relationships':
            return _RelationshipProxyResource(self.request, etype, parent=self)
        vreg = self.request.registry['cubicweb.registry']
        schema = vreg.schema
        try:
            rschema, __, role = find_relationship(schema, etype, value)
        except ValueError:
            raise KeyError(value)
        else:
            return RelatedEntitiesResource(
                self.request, rschema.type, role, parent=self)

    @reify
    def rset(self):
        req = self.request.cw_request
        if self.cls is None:
            return req.execute('Any X WHERE X eid %(x)s',
                               {'x': int(self.__name__)})
        st = self.cls.fetch_rqlst(self.request.cw_cnx.user, ordermethod=None)
        st.add_constant_restriction(st.get_variable('X'), self.attrname,
                                    'x', 'Substitute')
        if self.attrname == 'eid':
            try:
                rset = req.execute(st.as_string(), {'x': int(self.__name__)})
            except (ValueError, TypeResolverException):
                # conflicting eid/type
                raise HTTPNotFound()
        else:
            rset = req.execute(st.as_string(), {'x': text_type(self.__name__)})
        return rset


class RelatedEntitiesResource(object):
    """A resource wrapping entities related to the entity bound to
    `entity_resource` through `rtype`/`role`.
    """

    @parent
    def __init__(self, request, rtype, role, **kwargs):
        self.request = request
        self.rtype = rtype
        self.role = role
        self.__name__ = rtype

    @reify
    def rset(self):
        # May raise HTTPNotFound, this is probably fine (isn't it?)
        entity = self.__parent__.rset.one()
        vreg = self.request.registry['cubicweb.registry']
        # XXX Until https://www.cubicweb.org/ticket/12306543 gets done.
        rql = entity.cw_related_rql(self.rtype, role=self.role)
        args = {'x': entity.eid}
        select = vreg.parse(entity._cw, rql, args).children[0]
        sortterms = self.request.params.get('sort')
        if sortterms:
            select.remove_sort_terms()
            mainvar = select.get_variable('X')
            for term in sortterms.split(','):
                if term.startswith('-'):
                    asc = False
                    term = term[1:]
                else:
                    asc = True
                mdvar = select.make_variable()
                rel = nodes.make_relation(mainvar, term,
                                          (mdvar,), nodes.VariableRef)
                select.add_restriction(rel)
                select.add_sort_var(mdvar, asc=asc)
        rql = select.as_string()
        return entity._cw.execute(rql, args)

    def __getitem__(self, value):
        if self.role != 'subject':
            # We only handle subject-relation in traversal.
            raise KeyError(value)
        entity = self.__parent__.rset.one()
        rset = self.request.cw_cnx.execute(
            'Any O WHERE O eid %(o)s, S {rtype} O, S eid %(s)s'.format(
                rtype=self.rtype),
            {'o': value, 's': entity.eid})
        if not rset:
            raise KeyError(value)
        return RelatedEntityResource(
            self.request, rset.one(), rtype=self.rtype, role=self.role,
            parent=self)


class RelatedEntityResource(object):
    """A resource an entity related to another one."""

    def __init__(self, request, entity, rtype, role, parent):
        self.request = request
        self.entity = entity
        self.rtype = rtype
        self.role = role
        self.__parent__ = parent
        self.__name__ = text_type(entity.eid)



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-jsonschema views/forms/actions/components for web ui"""

from cubicweb.__pkginfo__ import numversion as cw_version
from cubicweb import neg_role
from cubicweb.rtags import RelationTags
from cubicweb.schema import VIRTUAL_RTYPES, WORKFLOW_RTYPES


class JSONSchemaSectionRelationTags(RelationTags):
    """JSON Schema configuration relation tag.

    This relation tag can be used to configure generation of JSON Schema
    documents through uicfg. It accepts the following sections:

    *   'inlined': relation definition would appear as an inlined document of
        the main entity's JSON Schema,
    *   'hidden': relation definition is hidden from JSON Schema, and,
    *   'related': relation definition would not appear in entity's JSON
        Schema but would appear in a rel="related" hypermedia link.
    """
    __regid__ = 'jsonschema'

    _allowed_values = frozenset(('inlined', 'hidden', 'related'))
    ignored_rtypes = VIRTUAL_RTYPES | WORKFLOW_RTYPES

    def _init(self, sschema, rschema, oschema, role):
        if self.get(sschema, rschema, oschema, role) is not None:
            return
        if (rschema.meta
                or sschema.is_metadata(rschema)
                or rschema.type in self.ignored_rtypes):
            section = 'hidden'
        elif rschema.final:
            # TODO Should we handle Password (oschema.type)? Maybe not because
            # we already have a mapper producing no value.
            section = 'inlined'
        else:
            rdef = rschema.rdef(sschema, oschema)
            if rdef.role_cardinality(role) in '1+':
                section = 'inlined'
            elif rdef.composite == neg_role(role):
                section = 'inlined'
            else:
                section = 'related'
        self.tag_relation((sschema, rschema, oschema, role), section)

    def relations_by_section(self, entity, section, permission):
        """return a list of (rtype, role, target types) for the given entity
        matching 'section' and 'permission'.
        """
        eschema = entity.e_schema
        cnx = entity._cw
        if permission == 'update':
            relpermission = 'add'
        else:
            relpermission = permission
        eid = entity.eid if entity.has_eid() else None
        rdefs = eschema.relation_definitions(includefinal=True)
        for rschema, targetschemas, role in rdefs:
            if rschema in self.ignored_rtypes:
                continue
            relperms_kwargs = {
                'subject': {'fromeid': eid},
                'object': {'toeid': eid},
            }[role]
            target_types = set()
            for tschema in targetschemas:
                if tschema.type == 'Password' and relpermission == 'read':
                    continue
                if section not in self.etype_get(eschema, rschema, role,
                                                 tschema):
                    continue
                rdef = rschema.role_rdef(eschema, tschema, role)
                if rschema.final:
                    if not rdef.has_perm(cnx, permission, eid=eid,
                                         creating=eid is None):
                        continue
                else:
                    if not rdef.has_perm(cnx, relpermission, **relperms_kwargs):
                        continue
                target_types.add(tschema.type)
            if target_types:
                yield rschema.type, role, target_types


if cw_version >= (3, 25):
    jsonschema_section = JSONSchemaSectionRelationTags(__module__=__name__)
else:
    jsonschema_section = JSONSchemaSectionRelationTags()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-jsonschema yams constraints to jsl serializers"""

from logilab.common.registry import Predicate

from . import JSONSchemaMapper


class yams_constraint_type(Predicate):
    """ Predicate that returns 1 when the yams constraint supplied at execution
    time matches the constraint type name supplied at construction time (e.g.
    "SizeConstraint", "StaticVocabularyConstraint", etc.).
    """

    def __init__(self, cstr_type):
        self.cstr_type = cstr_type

    def __call__(self, cls, vreg, _, etype, rtype, cstr, jsl_attrs):
        if cstr.type() == self.cstr_type:
            return 1
        return 0


class JslConstraintSerializer(JSONSchemaMapper):
    __regid__ = 'jsonschema.constraint'
    __abstract__ = True

    def __init__(self, vreg, _, etype, rtype, constraint, jsl_attrs):
        self.vreg = vreg
        self._ = _
        self.etype = vreg.schema[etype]
        self.rtype = vreg.schema[rtype]
        self.constraint = constraint
        self.jsl_attrs = jsl_attrs

    def todict(self):
        raise NotImplementedError


class JslSizeConstraintSerializer(JslConstraintSerializer):
    __select__ = yams_constraint_type('SizeConstraint')

    def todict(self):
        if ('min_length' not in self.jsl_attrs
                and self.constraint.min is not None):
            return {'min_length': self.constraint.min}
        if ('max_length' not in self.jsl_attrs
                and self.constraint.max is not None):
            return {'max_length': self.constraint.max}


class JslIntervalBoundConstraintSerializer(JslConstraintSerializer):
    __select__ = yams_constraint_type('IntervalBoundConstraint')

    def todict(self):
        if ('minimum' not in self.jsl_attrs
                and self.constraint.minvalue is not None):
            return {'minimum': self.constraint.minvalue}
        if ('maximum' not in self.jsl_attrs
                and self.constraint.maxvalue is not None):
            return {'maximum': self.constraint.maxvalue}
